""" 
Path tracking module.
    
author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

import numpy as np
cimport numpy as cnp


cpdef cnp.ndarray  compute_path(cnp.uint32_t[::1] path_links, int vertex_idx):
    
    cdef int path_length

    path_length = compute_path_first_pass(path_links, vertex_idx)
    path_vertices = np.empty(path_length, dtype=np.uint32)
    compute_path_second_pass(path_links, path_vertices, vertex_idx)

    return path_vertices
    

cdef int compute_path_first_pass(
    cnp.uint32_t[::1] path_links,  
    int vertex_idx
) nogil:
    """Returns the path length.
    """

    cdef:
        size_t i, j
        int k

    # initialization
    j = <size_t>vertex_idx
    i = j + 1
    k = 0

    # loop
    while i != j:
        i = j
        j = <size_t>path_links[j]
        k += 1


    return k

cdef void compute_path_second_pass(
    cnp.uint32_t[::1] path_links,
    cnp.uint32_t[::1] path_vertices,  
    int vertex_idx
) nogil:

    cdef size_t i, j, k

    # initialization
    j = <size_t>vertex_idx
    i = j + 1
    k = 0

    # loop
    while i != j:
        i = j
        path_vertices[k] = j 
        j = <size_t>path_links[j]
        k += 1


""" 
Common definitions. Header file.

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

cimport numpy as cnp

# priority queue
# --------------

ctypedef cnp.float64_t DTYPE_t
cdef DTYPE_t DTYPE_INF
cdef DTYPE_t INF_FREQ
cdef DTYPE_t MIN_FREQ 
cdef DTYPE_t A_VERY_SMALL_TIME_INTERVAL

cdef enum ElementState:
   SCANNED
   UNLABELED
   LABELED


""" Priority queue based on a minimum binary heap. Header file.

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

cimport numpy as cnp

from edsger.commons cimport DTYPE_t, ElementState


cdef struct Element:
    DTYPE_t key
    ElementState state
    size_t node_idx

cdef struct PriorityQueue:
    size_t length  # number of elements in the array
    size_t size  # number of elements in the heap
    size_t* A  # array storing the binary tree
    Element* Elements  # array storing the elements

cdef void init_pqueue(PriorityQueue*, size_t, size_t) nogil
cdef void free_pqueue(PriorityQueue*) nogil
cdef void insert(PriorityQueue*, size_t, DTYPE_t) nogil
cdef DTYPE_t peek(PriorityQueue*) nogil
cdef size_t extract_min(PriorityQueue*) nogil
cdef bint is_empty(PriorityQueue*) nogil
cdef void decrease_key(PriorityQueue*, size_t, DTYPE_t) nogil
cdef cnp.ndarray copy_keys_to_numpy(PriorityQueue*, size_t)



""" 
Common definitions.

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

import numpy as np

DTYPE_PY = np.float64
DTYPE_INF = <DTYPE_t>np.finfo(dtype=DTYPE_PY).max
DTYPE_INF_PY = DTYPE_INF

# Spiess & Florian
# ----------------

# infinite frequency is defined here numerically
# this must be a very large number depending on the precision on the computation
# INF_FREQ << DTYPE_INF
INF_FREQ = 1.0e+20
INF_FREQ_PY = INF_FREQ

# smallest frequency
# WARNING: this must be small but not too small 
# 1 / MIN_FREQ << DTYPE_INF
MIN_FREQ = 1.0 / INF_FREQ
MIN_FREQ_PY =  MIN_FREQ

# a very small time interval
A_VERY_SMALL_TIME_INTERVAL = 1.0e+08 * MIN_FREQ
A_VERY_SMALL_TIME_INTERVAL_PY = A_VERY_SMALL_TIME_INTERVAL



""" Module dedicated to the generation of small networks, for testing purposes.
"""

import numpy as np
import pandas as pd


def create_SF_network(dwell_time=1.0e-6, board_alight_ratio=0.5):
    """
    Example network from Spiess, H. and Florian, M. (1989).
    Optimal strategies: A new assignment model for transit networks.
    Transportation Research Part B 23(2), 83-102.

    This network has 13 vertices and 24 edges.
    """

    boarding_time = board_alight_ratio * dwell_time
    alighting_time = board_alight_ratio * dwell_time

    line1_freq = 1.0 / (60.0 * 12.0)
    line2_freq = 1.0 / (60.0 * 12.0)
    line3_freq = 1.0 / (60.0 * 30.0)
    line4_freq = 1.0 / (60.0 * 6.0)

    # stop A
    # 0 stop vertex
    # 1 boarding vertex
    # 2 boarding vertex

    # stop X
    # 3 stop vertex
    # 4 boarding vertex
    # 5 alighting vertex
    # 6 boarding vertex

    # stop Y
    # 7  stop vertex
    # 8  boarding vertex
    # 9  alighting vertex
    # 10 boarding vertex
    # 11 alighting vertex

    # stop B
    # 12 stop vertex
    # 13 alighting vertex
    # 14 alighting vertex
    # 15 alighting vertex

    tail = []
    head = []
    trav_time = []
    freq = []
    vol = []

    # edge 0
    # stop A : to line 1
    # boarding edge
    tail.append(0)
    head.append(2)
    freq.append(line1_freq)
    trav_time.append(boarding_time)
    vol.append(0.5)

    # edge 1
    # stop A : to line 2
    # boarding edge
    tail.append(0)
    head.append(1)
    freq.append(line2_freq)
    trav_time.append(boarding_time)
    vol.append(0.5)

    # edge 2
    # line 1 : first segment
    # on-board edge
    tail.append(2)
    head.append(15)
    freq.append(np.inf)
    trav_time.append(25.0 * 60.0)
    vol.append(0.5)

    # edge 3
    # line 2 : first segment
    # on-board edge
    tail.append(1)
    head.append(5)
    freq.append(np.inf)
    trav_time.append(7.0 * 60.0)
    vol.append(0.5)

    # edge 4
    # stop X : from line 2
    # alighting edge
    tail.append(5)
    head.append(3)
    freq.append(np.inf)
    trav_time.append(alighting_time)
    vol.append(0.0)

    # edge 5
    # stop X : in line 2
    # dwell edge
    tail.append(5)
    head.append(6)
    freq.append(np.inf)
    trav_time.append(dwell_time)
    vol.append(0.5)

    # edge 6
    # stop X : from line 2 to line 3
    # transfer edge
    tail.append(5)
    head.append(4)
    freq.append(line3_freq)
    trav_time.append(dwell_time)
    vol.append(0.0)

    # edge 7
    # stop X : to line 2
    # boarding edge
    tail.append(3)
    head.append(6)
    freq.append(line2_freq)
    trav_time.append(boarding_time)
    vol.append(0.0)

    # edge 8
    # stop X : to line 3
    # boarding edge
    tail.append(3)
    head.append(4)
    freq.append(line3_freq)
    trav_time.append(boarding_time)
    vol.append(0.0)

    # edge 9
    # line 2 : second segment
    # on-board edge
    tail.append(6)
    head.append(11)
    freq.append(np.inf)
    trav_time.append(6.0 * 60.0)
    vol.append(0.5)

    # edge 10
    # line 3 : first segment
    # on-board edge
    tail.append(4)
    head.append(9)
    freq.append(np.inf)
    trav_time.append(4.0 * 60.0)
    vol.append(0.0)

    # edge 11
    # stop Y : from line 3
    # alighting edge
    tail.append(9)
    head.append(7)
    freq.append(np.inf)
    trav_time.append(alighting_time)
    vol.append(0.0)

    # edge 12
    # stop Y : from line 2
    # alighting edge
    tail.append(11)
    head.append(7)
    freq.append(np.inf)
    trav_time.append(alighting_time)
    vol.append(0.0)

    # edge 13
    # stop Y : from line 2 to line 3
    # transfer edge
    tail.append(11)
    head.append(10)
    freq.append(line3_freq)
    trav_time.append(dwell_time)
    vol.append(0.0833333333333)

    # edge 14
    # stop Y : from line 2 to line 4
    # transfer edge
    tail.append(11)
    head.append(8)
    freq.append(line4_freq)
    trav_time.append(dwell_time)
    vol.append(0.4166666666666)

    # edge 15
    # stop Y : from line 3 to line 4
    # transfer edge
    tail.append(9)
    head.append(8)
    freq.append(line4_freq)
    trav_time.append(dwell_time)
    vol.append(0.0)

    # edge 16
    # stop Y : in line 3
    # dwell edge
    tail.append(9)
    head.append(10)
    freq.append(np.inf)
    trav_time.append(dwell_time)
    vol.append(0.0)

    # edge 17
    # stop Y : to line 3
    # boarding edge
    tail.append(7)
    head.append(10)
    freq.append(line3_freq)
    trav_time.append(boarding_time)
    vol.append(0.0)

    # edge 18
    # stop Y : to line 4
    # boarding edge
    tail.append(7)
    head.append(8)
    freq.append(line4_freq)
    trav_time.append(boarding_time)
    vol.append(0.0)

    # edge 19
    # line 3 : second segment
    # on-board edge
    tail.append(10)
    head.append(14)
    freq.append(np.inf)
    trav_time.append(4.0 * 60.0)
    vol.append(0.0833333333333)

    # edge 20
    # line 4 : first segment
    # on-board edge
    tail.append(8)
    head.append(13)
    freq.append(np.inf)
    trav_time.append(10.0 * 60.0)
    vol.append(0.4166666666666)

    # edge 21
    # stop Y : from line 1
    # alighting edge
    tail.append(15)
    head.append(12)
    freq.append(np.inf)
    trav_time.append(alighting_time)
    vol.append(0.5)

    # edge 22
    # stop Y : from line 3
    # alighting edge
    tail.append(14)
    head.append(12)
    freq.append(np.inf)
    trav_time.append(alighting_time)
    vol.append(0.0833333333333)

    # edge 23
    # stop Y : from line 4
    # alighting edge
    tail.append(13)
    head.append(12)
    freq.append(np.inf)
    trav_time.append(alighting_time)
    vol.append(0.4166666666666)

    edges = pd.DataFrame(
        data={
            "tail": tail,
            "head": head,
            "trav_time": trav_time,
            "freq": freq,
            "volume_ref": vol,
        }
    )
    # waiting time is in average half of the period
    edges["freq"] *= 2.0

    return edges



""" Tests for pq_bin_dec_0b.pyx.

py.test tests/test_pq_4ary_dec_0b.py

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

from edsger.pq_4ary_dec_0b import *


def test_init_01():
    init_01()


def test_insert_01():
    insert_01()


def test_insert_02():
    insert_02()


def test_insert_03():
    insert_03()


def test_peek_01():
    peek_01()


def test_extract_min_01():
    extract_min_01()


def test_is_empty_01():
    is_empty_01()


def test_decrease_key_01():
    decrease_key_01()


def test_sort_01():
    sort_01(1000)



""" Tests for path_tracking.pyx.

py.test tests/test_path_tracking.py

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

import numpy as np

from edsger.path_tracking import compute_path


def test_01(n=10):
    path_links = np.arange(n, dtype=np.uint32)
    path_links[1:] -= 1

    path_vertices = compute_path(path_links, n - 1)
    path_vertices_ref = np.array(list(range(n))[::-1], dtype=np.uint32)

    np.testing.assert_array_equal(path_vertices, path_vertices_ref)



""" Tests for dijkstra.pyx.

py.test tests/test_dijkstra.py

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

from edsger.dijkstra import *


def test_compute_sssp_01():
    compute_sssp_01()


def test_compute_stsp_01():
    compute_stsp_01()


def test_compute_sssp_02():
    compute_sssp_02()


def test_compute_stsp_02():
    compute_stsp_02()



""" Tests for path.py.

py.test tests/test_path.py

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

import numpy as np
import pandas as pd
import pytest
from scipy.sparse import coo_array, csr_matrix
from scipy.sparse.csgraph import dijkstra

from edsger.commons import DTYPE_INF_PY
from edsger.networks import create_SF_network
from edsger.path import Dijkstra, HyperpathGenerating
from edsger.commons import A_VERY_SMALL_TIME_INTERVAL_PY


@pytest.fixture
def braess():
    """Braess-like graph"""
    edges = pd.DataFrame(
        data={
            "tail": [0, 0, 1, 1, 2],
            "head": [1, 2, 2, 3, 3],
            "weight": [1.0, 2.0, 0.0, 2.0, 1.0],
        }
    )
    return edges


def test_check_edges_01():
    """Negative weights."""

    edges = pd.DataFrame(
        data={
            "tail": [0, 0],
            "head": [1, 2],
            "weight": [1.0, -2.0],
        }
    )
    with pytest.raises(ValueError, match=r"nonnegative"):
        sp = Dijkstra(
            edges,
            check_edges=True,
        )


def test_check_edges_02(braess):
    edges = braess

    with pytest.raises(TypeError, match=r"pandas DataFrame"):
        sp = Dijkstra("yeaaahhh!!!", check_edges=True)
    with pytest.raises(KeyError, match=r"not found in graph edges dataframe"):
        sp = Dijkstra(edges, tail="source", check_edges=True)
    with pytest.raises(KeyError, match=r"not found in graph edges dataframe"):
        sp = Dijkstra(edges, head="target", check_edges=True)
    with pytest.raises(KeyError, match=r"not found in graph edges dataframe"):
        sp = Dijkstra(edges, weight="cost", check_edges=True)
    with pytest.raises(ValueError, match=r"missing value"):
        sp = Dijkstra(edges.replace(0, np.nan), check_edges=True)
    with pytest.raises(TypeError, match=r"should be of integer type"):
        sp = Dijkstra(edges.astype({"tail": float}), check_edges=True)
    with pytest.raises(TypeError, match=r"should be of numeric type"):
        sp = Dijkstra(edges.astype({"weight": str}), check_edges=True)


def test_run_01(braess):
    edges = braess
    sp = Dijkstra(edges, orientation="out", check_edges=False)
    path_lengths = sp.run(vertex_idx=0, return_Series=True)
    path_lengths_ref = pd.Series([0.0, 1.0, 1.0, 2.0])
    path_lengths_ref.index.name = "vertex_idx"
    path_lengths_ref.name = "path_length"
    pd.testing.assert_series_equal(path_lengths, path_lengths_ref)


def test_run_02(random_seed=124, n=1000):
    np.random.seed(random_seed)
    tail = np.random.randint(0, int(n / 5), n)
    head = np.random.randint(0, int(n / 5), n)
    weight = np.random.rand(n)
    edges = pd.DataFrame(data={"tail": tail, "head": head, "weight": weight})
    edges.drop_duplicates(subset=["tail", "head"], inplace=True)
    edges = edges.loc[edges["tail"] != edges["head"]]
    edges.reset_index(drop=True, inplace=True)

    # SciPy
    vertex_count = edges[["tail", "head"]].max().max() + 1
    data = edges["weight"].values
    row = edges["tail"].values
    col = edges["head"].values
    graph_coo = coo_array((data, (row, col)), shape=(vertex_count, vertex_count))
    graph_csr = graph_coo.tocsr()
    dist_matrix_ref = dijkstra(
        csgraph=graph_csr, directed=True, indices=0, return_predecessors=False
    )

    # In-house
    # without graph permutation
    # return_inf=True
    sp = Dijkstra(edges, orientation="out", check_edges=True, permute=False)
    dist_matrix = sp.run(vertex_idx=0, return_inf=True)
    assert np.allclose(dist_matrix, dist_matrix_ref)

    dist_matrix_ref = np.where(
        dist_matrix_ref > DTYPE_INF_PY, DTYPE_INF_PY, dist_matrix_ref
    )

    # without graph permutation
    # return_inf=False
    dist_matrix = sp.run(vertex_idx=0, return_inf=False)
    assert np.allclose(dist_matrix, dist_matrix_ref)

    # with graph permutation
    # return_inf=False
    sp = Dijkstra(edges, orientation="out", check_edges=True, permute=True)
    dist_matrix = sp.run(vertex_idx=0, return_inf=False)
    assert np.allclose(dist_matrix, dist_matrix_ref)


def test_run_03(braess):
    """
    orientation="in"
    """
    edges = braess
    sp = Dijkstra(edges, orientation="in")
    path_lengths = sp.run(vertex_idx=3)
    path_lengths_ref = [2.0, 1.0, 1.0, 0.0]
    assert np.allclose(path_lengths, path_lengths_ref)


def test_run_04():
    """
    permute=True
    """
    edges = pd.DataFrame(
        data={
            "tail": [0, 0, 10, 10, 20],
            "head": [10, 20, 20, 30, 30],
            "weight": [1.0, 2.0, 0.0, 2.0, 1.0],
        }
    )
    sp = Dijkstra(edges, orientation="out", permute=True)
    path_lengths = sp.run(
        vertex_idx=0, path_tracking=True, return_inf=True, return_Series=False
    )
    path_lengths_ref = np.array(
        [
            0.0,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            1.0,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            1.0,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            np.inf,
            2.0,
        ]
    )
    assert np.allclose(path_lengths, path_lengths_ref)
    path_links_ref = np.array(
        [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            0,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            10,
            21,
            22,
            23,
            24,
            25,
            26,
            27,
            28,
            29,
            20,
        ],
        dtype=np.uint32,
    )
    np.testing.assert_array_equal(sp._path_links, path_links_ref)

    path_lengths = sp.run(
        vertex_idx=0, path_tracking=True, return_inf=False, return_Series=False
    )
    path_lengths_ref = np.array(
        [
            0.00000000e000,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            1.00000000e000,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            1.00000000e000,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            DTYPE_INF_PY,
            2.00000000e000,
        ]
    )
    assert np.allclose(path_lengths, path_lengths_ref)
    np.testing.assert_array_equal(sp._path_links, path_links_ref)

    path_lengths = sp.run(vertex_idx=0, path_tracking=True, return_Series=True)
    path_lengths_ref = pd.Series([0.0, 1.0, 1.0, 2.0], index=[0, 10, 20, 30])
    path_lengths_ref.index.name = "vertex_idx"
    path_lengths_ref.name = "path_length"
    pd.testing.assert_series_equal(path_lengths, path_lengths_ref)
    path_links_ref = pd.Series([0, 0, 10, 20], index=[0, 10, 20, 30])
    path_links_ref.index.name = "vertex_idx"
    path_links_ref.name = "associated_idx"
    path_links_ref = path_links_ref.astype(np.uint32)
    np.testing.assert_array_equal(sp._path_links, path_links_ref)


def test_path_tracking_01():
    edges = pd.DataFrame(
        data={
            "tail": [0, 0, 10, 10, 20],
            "head": [10, 20, 20, 30, 30],
            "weight": [1.0, 2.0, 0.0, 2.0, 1.0],
        }
    )
    sp = Dijkstra(edges, orientation="out", permute=True)

    # trying to get the path while no shortest path alrgorithm
    # has been executed yet
    with pytest.warns(UserWarning) as record:
        sp.get_path(0)

    # check that only one warning was raised
    assert len(record) == 1

    # run the shortest path algorithm
    path_lengths = sp.run(
        vertex_idx=0, path_tracking=True, return_inf=True, return_Series=False
    )
    path_links_ref = np.array(
        [
            0,
            1,
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            0,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            10,
            21,
            22,
            23,
            24,
            25,
            26,
            27,
            28,
            29,
            20,
        ],
        dtype=np.uint32,
    )
    np.testing.assert_array_equal(sp._path_links, path_links_ref)

    path_vertices = sp.get_path(30)
    path_vertices_ref = np.array([30, 20, 10, 0], dtype=np.uint32)
    np.testing.assert_array_equal(path_vertices, path_vertices_ref)


def test_SF_in_01():
    edges = create_SF_network(dwell_time=0.0)
    hp = HyperpathGenerating(edges, check_edges=False)
    hp.run(origin=0, destination=12, volume=1.0)

    assert np.allclose(edges["volume_ref"].values, hp._edges["volume"].values)

    u_i_vec_ref = np.array(
        [
            1.66500000e03,
            1.47000000e03,
            1.50000000e03,
            1.14428572e03,
            4.80000000e02,
            1.05000000e03,
            1.05000000e03,
            6.90000000e02,
            6.00000000e02,
            2.40000000e02,
            2.40000000e02,
            6.90000000e02,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
            0.00000000e00,
        ]
    )
    assert np.allclose(u_i_vec_ref, hp.u_i_vec, rtol=1e-08, atol=1e-08)


def test_SF_dwell_and_transfer_01():
    """
    One line, one stop case. We look at what is happening at a given stop.
    Do people use the dwell edge (what they should do) or the alighting and
    boarding edges.

            0
      0 --------> 2
        \\      ^
        1\\    /2
          \\  /
           V /
            1

    Network with 3 vertices and 3 edges.
    """

    # line 1
    line_freq = 1.0 / 600.0
    dwell_time = 5.0

    # stop A
    # 0 alighting vertex
    # 1 stop vertex
    # 2 boarding vertex

    tail = []
    head = []
    trav_time = []
    freq = []
    vol = []

    # edge 0
    # stop A : in line 1
    # dwell edge
    tail.append(0)
    head.append(2)
    freq.append(np.inf)
    trav_time.append(dwell_time)
    vol.append(1.0)

    # edge 1
    # stop A : from line 1
    # alighting edge
    tail.append(0)
    head.append(1)
    freq.append(np.inf)
    trav_time.append(0.5 * dwell_time)
    vol.append(0.0)

    # edge 2
    # stop A : to line 1
    # boarding edge
    tail.append(1)
    head.append(2)
    freq.append(line_freq)
    trav_time.append(0.5 * dwell_time)
    vol.append(0.0)

    edges = pd.DataFrame(
        data={
            "tail": tail,
            "head": head,
            "trav_time": trav_time,
            "freq": freq,
            "volume_ref": vol,
        }
    )
    # waiting time is in average half of the period
    edges["freq"] *= 2.0

    # SF
    hp = HyperpathGenerating(edges, check_edges=False)
    hp.run(origin=0, destination=2, volume=1.0)

    assert np.allclose(edges["volume_ref"].values, hp._edges["volume"].values)

    u_i_vec_ref = np.array(
        [
            dwell_time,
            0.5 * (1.0 / line_freq + dwell_time),
            0.0,
        ]
    )
    assert np.allclose(u_i_vec_ref, hp.u_i_vec, rtol=1e-08, atol=1e-08)

    # now we change the dwell edge into a transfer edge
    freq[0] = line_freq

    # vol = 3*[0.5]
    edges = pd.DataFrame(
        data={
            "tail": tail,
            "head": head,
            "trav_time": trav_time,
            "freq": freq,
            "volume_ref": vol,
        }
    )
    # waiting time is in average half of the period
    edges["freq"] *= 2.0

    # SF
    hp = HyperpathGenerating(edges, check_edges=False)
    hp.run(origin=0, destination=2, volume=1.0)

    # edges 1 and 2 are not used because edge 1 does not have
    # a real infinite frequency but INF_FREQ value, for numerical reasons.
    # This implies a small resistance to the path going through vertex 1.
    # If we decrease by a tiny amount the frequency of edge 0, the flow goes
    # though vertex 1.
    assert np.allclose(edges["volume_ref"].values, hp._edges["volume"].values)

    u_i_vec_ref = [305.0, 302.5, 0.0]
    assert np.allclose(u_i_vec_ref, hp.u_i_vec, rtol=1e-08, atol=1e-08)



""" 
Priority queue based on a minimum 4-ary heap.

- dec: priority queue with decrease-key operation.
- 0b: indices are 0-based
    
author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

cimport numpy as cnp
from libc.stdlib cimport free, malloc

from edsger.commons import DTYPE_PY
from edsger.commons cimport (
    DTYPE_INF, LABELED, UNLABELED, SCANNED, DTYPE_t
)


cdef void init_pqueue(
    PriorityQueue* pqueue,
    size_t heap_length,
    size_t element_count) nogil:
    """
    Initialize the priority queue.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * size_t length : length (maximum size) of the heap
    * size_t element_count : element count

    assumptions
    ===========
    * heap_length <= element_count
    """
    cdef size_t i

    pqueue.length = heap_length
    pqueue.size = 0
    pqueue.A = <size_t*> malloc(heap_length * sizeof(size_t))
    pqueue.Elements = <Element*> malloc(element_count * sizeof(Element))

    for i in range(heap_length):
        pqueue.A[i] = heap_length
        _initialize_element(pqueue, i)
    for i in range(heap_length, element_count):
        _initialize_element(pqueue, i)


cdef void _initialize_element(
    PriorityQueue* pqueue,
    size_t element_idx) nogil:
    """
    Initialize a single element.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * size_t element_idx : index of the element in the element array
    """
    pqueue.Elements[element_idx].key = DTYPE_INF
    pqueue.Elements[element_idx].state = UNLABELED
    pqueue.Elements[element_idx].node_idx = pqueue.length


cdef void free_pqueue(
    PriorityQueue* pqueue) nogil:
    """
    Free the priority queue.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    """
    free(pqueue.A)
    free(pqueue.Elements)


cdef void insert(
    PriorityQueue* pqueue,
    size_t element_idx,
    DTYPE_t key) nogil:
    """
    Insert an element into the heap and reorder the heap.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * size_t element_idx : index of the element in the element array
    * DTYPE_t key : key value of the element

    assumptions
    ===========
    * the element pqueue.Elements[element_idx] is not in the heap
    * its new key is smaller than DTYPE_INF
    """
    cdef size_t node_idx = pqueue.size

    pqueue.size += 1
    pqueue.Elements[element_idx].state = LABELED
    pqueue.Elements[element_idx].node_idx = node_idx
    pqueue.A[node_idx] = element_idx
    _decrease_key_from_node_index(pqueue, node_idx, key)


cdef void decrease_key(
    PriorityQueue* pqueue, 
    size_t element_idx, 
    DTYPE_t key_new) nogil:
    """
    Decrease the key of a element in the heap, given its element index.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * size_t element_idx : index of the element in the element array
    * DTYPE_t key_new : new value of the element key 

    assumption
    ==========
    * pqueue.Elements[idx] is in the heap
    """
    _decrease_key_from_node_index(
        pqueue, 
        pqueue.Elements[element_idx].node_idx, 
        key_new)


cdef DTYPE_t peek(PriorityQueue* pqueue) nogil:
    """
    Find heap min key.

    input
    =====
    * PriorityQueue* pqueue : priority queue

    output
    ======
    * DTYPE_t : key value of the min element

    assumption
    ==========
    * pqueue.size > 0
    * heap is heapified
    """
    return pqueue.Elements[pqueue.A[0]].key


cdef bint is_empty(PriorityQueue* pqueue) nogil:
    """
    Check if the heap is empty.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    """
    cdef bint isempty = 0

    if pqueue.size == 0:
        isempty = 1

    return isempty


cdef size_t extract_min(PriorityQueue* pqueue) nogil:
    """
    Extract element with min keay from the heap, 
    and return its element index.

    input
    =====
    * PriorityQueue* pqueue : priority queue

    output
    ======
    * size_t : element index with min key

    assumption
    ==========
    * pqueue.size > 0
    """
    cdef: 
        size_t element_idx = pqueue.A[0]  # min element index
        size_t node_idx = pqueue.size - 1  # last leaf node index

    # exchange the root node with the last leaf node
    _exchange_nodes(pqueue, 0, node_idx)

    # remove this element from the heap
    pqueue.Elements[element_idx].state = SCANNED
    pqueue.Elements[element_idx].node_idx = pqueue.length
    pqueue.A[node_idx] = pqueue.length
    pqueue.size -= 1

    # reorder the tree Elements from the root node
    _min_heapify(pqueue, 0)

    return element_idx

cdef void _exchange_nodes(
    PriorityQueue* pqueue, 
    size_t node_i,
    size_t node_j) nogil:
    """
    Exchange two nodes in the heap.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * size_t node_i: first node index
    * size_t node_j: second node index
    """
    cdef: 
        size_t element_i = pqueue.A[node_i]
        size_t element_j = pqueue.A[node_j]
    
    # exchange element indices in the heap array
    pqueue.A[node_i] = element_j
    pqueue.A[node_j] = element_i

    # exchange node indices in the element array
    pqueue.Elements[element_j].node_idx = node_i
    pqueue.Elements[element_i].node_idx = node_j


cdef void _min_heapify(
    PriorityQueue* pqueue,
    size_t node_idx) nogil:
    """
    Re-order sub-tree under a given node (given its node index) 
    until it satisfies the heap property.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * size_t node_idx : node index
    """
    cdef: 
        size_t c1, c2, c3, c4, i = node_idx, s
        DTYPE_t val_tmp, val_min

    while True:

        c1 = 4 * i + 1  
        c2 = c1 + 1
        c3 = c2 + 1
        c4 = c3 + 1
        
        s = i
        val_min = pqueue.Elements[pqueue.A[s]].key
        if (c4 < pqueue.size):
            val_tmp = pqueue.Elements[pqueue.A[c4]].key
            if val_tmp < val_min:
                s = c4
                val_min = val_tmp
            val_tmp = pqueue.Elements[pqueue.A[c3]].key
            if val_tmp < val_min:
                s = c3
                val_min = val_tmp
            val_tmp = pqueue.Elements[pqueue.A[c2]].key
            if val_tmp < val_min:
                s = c2
                val_min = val_tmp
            val_tmp = pqueue.Elements[pqueue.A[c1]].key
            if val_tmp < val_min:
                s = c1
        else:
            if (c3 < pqueue.size):
                val_tmp = pqueue.Elements[pqueue.A[c3]].key
                if val_tmp < val_min:
                    s = c3
                    val_min = val_tmp
                val_tmp = pqueue.Elements[pqueue.A[c2]].key
                if val_tmp < val_min:
                    s = c2
                    val_min = val_tmp
                val_tmp = pqueue.Elements[pqueue.A[c1]].key
                if val_tmp < val_min:
                    s = c1
            else:
                if (c2 < pqueue.size):
                    val_tmp = pqueue.Elements[pqueue.A[c2]].key
                    if val_tmp < val_min:
                        s = c2
                        val_min = val_tmp
                    val_tmp = pqueue.Elements[pqueue.A[c1]].key
                    if val_tmp < val_min:
                        s = c1
                else:
                    if (c1 < pqueue.size):
                        val_tmp = pqueue.Elements[pqueue.A[c1]].key
                        if val_tmp < val_min:
                            s = c1

        if s != i:
            _exchange_nodes(pqueue, i, s)
            i = s
        else:
            break


cdef void _decrease_key_from_node_index(
    PriorityQueue* pqueue,
    size_t node_idx, 
    DTYPE_t key_new) nogil:
    """
    Decrease the key of an element in the heap, given its tree index.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * size_t node_idx : node index
    * DTYPE_t key_new : new key value

    assumptions
    ===========
    * pqueue.Elements[pqueue.A[node_idx]] is in the heap (node_idx < pqueue.size)
    * key_new < pqueue.Elements[pqueue.A[node_idx]].key
    """
    cdef:
        size_t i = node_idx, j
        DTYPE_t key_j

    pqueue.Elements[pqueue.A[i]].key = key_new
    while i > 0: 
        j = (i - 1) // 4
        key_j = pqueue.Elements[pqueue.A[j]].key
        if key_j > key_new:
            _exchange_nodes(pqueue, i, j)
            i = j
        else:
            break

cdef cnp.ndarray copy_keys_to_numpy(
    PriorityQueue* pqueue,
    size_t vertex_count
):
    """
    Copy the keys into a numpy array.

    input
    =====
    * PriorityQueue* pqueue : priority queue
    * int vertex_count : vertex count
    * int num_threads : number of threads for the parallel job

    output
    ======
    * cnp.ndarray : NumPy array with all the keys
    """

    path_lengths = cnp.ndarray(vertex_count, dtype=DTYPE_PY)

    cdef:
        size_t i  # loop counter
        DTYPE_t[::1] path_lengths_view = path_lengths

    with nogil:

        for i in range(vertex_count):
            path_lengths_view[i] = pqueue.Elements[i].key

    return path_lengths


# ============================================================================ #
# tests                                                                        #
# ============================================================================ #

import numpy as np


cpdef init_01(int length=4):
    """ 
    Initialize an empty priority queue.
    """

    cdef: 
        PriorityQueue pqueue
        size_t l = <size_t>length

    init_pqueue(&pqueue, l, l)

    assert pqueue.length == l
    assert pqueue.size == 0
    for i in range(l):
        assert pqueue.A[i] == pqueue.length
        assert pqueue.Elements[i].key == DTYPE_INF
        assert pqueue.Elements[i].state == UNLABELED
        assert pqueue.Elements[i].node_idx == pqueue.length

    free_pqueue(&pqueue)


cpdef insert_01():
    """ 
    Insert an element into an empty priority queue of length 1.
    """

    cdef: 
        PriorityQueue pqueue
        DTYPE_t key

    init_pqueue(&pqueue, 1, 1)
    assert pqueue.length == 1
    key = 1.0
    insert(&pqueue, 0, key)
    assert pqueue.size == 1
    assert pqueue.A[0] == 0
    assert pqueue.Elements[0].key == key
    assert pqueue.Elements[0].state == LABELED
    assert pqueue.Elements[0].node_idx == 0

    free_pqueue(&pqueue)


cpdef insert_02():
    """ 
    Insert 4 elements into an empty priority queue of length 4.
    """

    cdef: 
        PriorityQueue pqueue
        DTYPE_t key

    init_pqueue(&pqueue, 4, 4)

    elem_idx = 1
    key = 3.0
    insert(&pqueue, elem_idx, key)
    A_ref = [1, 4, 4, 4]
    for i in range(4):
        assert pqueue.A[i] == A_ref[i]
    assert pqueue.Elements[elem_idx].key == key
    assert pqueue.Elements[elem_idx].state == LABELED
    assert pqueue.Elements[1].node_idx == 0
    assert pqueue.size == 1

    elem_idx = 0
    key = 2.0
    insert(&pqueue, elem_idx, key)
    A_ref = [0, 1, 4, 4]
    for i in range(4):
        assert pqueue.A[i] == A_ref[i]
    assert pqueue.Elements[elem_idx].key == key
    assert pqueue.Elements[elem_idx].state == LABELED
    assert pqueue.Elements[0].node_idx == 0
    assert pqueue.Elements[1].node_idx == 1
    assert pqueue.size == 2

    elem_idx = 3
    key = 4.0
    insert(&pqueue, elem_idx, key)
    A_ref = [0, 1, 3, 4]
    for i in range(4):
        assert pqueue.A[i] == A_ref[i]
    assert pqueue.Elements[elem_idx].key == key
    assert pqueue.Elements[elem_idx].state == LABELED
    assert pqueue.Elements[0].node_idx == 0
    assert pqueue.Elements[1].node_idx == 1
    assert pqueue.Elements[3].node_idx == 2
    assert pqueue.size == 3

    elem_idx = 2
    key = 1.0
    insert(&pqueue, elem_idx, key)
    A_ref = [2, 1, 3, 0]
    for i in range(4):
        assert pqueue.A[i] == A_ref[i]
    assert pqueue.Elements[2].key == key
    assert pqueue.Elements[2].state == LABELED
    assert pqueue.Elements[0].node_idx == 3
    assert pqueue.Elements[1].node_idx == 1
    assert pqueue.Elements[2].node_idx == 0
    assert pqueue.Elements[3].node_idx == 2
    assert pqueue.size == 4

    free_pqueue(&pqueue)


cpdef insert_03(int length=4):
    """ 
    Insert elements with equal key values.
    """

    cdef: 
        PriorityQueue pqueue
        size_t i, l = <size_t>length
        DTYPE_t key = 1.0

    init_pqueue(&pqueue, l, l)
    for i in range(l):
        insert(&pqueue, i, key)
    for i in range(l):
        assert pqueue.A[i] == i

    free_pqueue(&pqueue)


cpdef peek_01():
    """ 
    Successively insert elements into a priority queue and
    check the min element of the priority without extracting it.
    """

    cdef PriorityQueue pqueue

    init_pqueue(&pqueue, 4, 4)

    insert(&pqueue, 0, 9.0)
    assert peek(&pqueue) == 9.0
    insert(&pqueue, 1, 9.0)
    assert peek(&pqueue) == 9.0
    insert(&pqueue, 2, 5.0)
    assert peek(&pqueue) == 5.0
    insert(&pqueue, 3, 3.0)
    assert peek(&pqueue) == 3.0

    free_pqueue(&pqueue)


cpdef extract_min_01():
    """ 
    Insert 4 elements into a priority queue and 
    extract them all.
    """
    
    cdef PriorityQueue pqueue

    init_pqueue(&pqueue, 4, 4)
    insert(&pqueue, 1, 3.0)
    insert(&pqueue, 0, 2.0)
    insert(&pqueue, 3, 4.0)
    insert(&pqueue, 2, 1.0)
    idx = extract_min(&pqueue)
    assert idx == 2
    assert pqueue.size == 3
    assert pqueue.Elements[idx].state == SCANNED
    idx = extract_min(&pqueue)
    assert idx == 0
    assert pqueue.size == 2
    assert pqueue.Elements[idx].state == SCANNED
    idx = extract_min(&pqueue)
    assert idx == 1
    assert pqueue.size == 1
    assert pqueue.Elements[idx].state == SCANNED
    idx = extract_min(&pqueue)
    assert idx == 3
    assert pqueue.size == 0
    assert pqueue.Elements[idx].state == SCANNED

    free_pqueue(&pqueue)


cpdef is_empty_01():
    """ 
    Insert an element and extract it.
    """
    
    cdef PriorityQueue pqueue

    init_pqueue(&pqueue, 2, 2)

    assert is_empty(&pqueue) == 1
    insert(&pqueue, 1, 3.0)
    assert is_empty(&pqueue) == 0
    idx = extract_min(&pqueue)
    assert is_empty(&pqueue) == 1

    free_pqueue(&pqueue)


cpdef decrease_key_01():
    """ 
    Insert elements into a priority queue and decrease the largest 
    key value to become the smallest.
    """

    cdef PriorityQueue pqueue

    init_pqueue(&pqueue, 4, 4)

    insert(&pqueue, 1, 3.0)
    insert(&pqueue, 0, 2.0)
    insert(&pqueue, 3, 4.0)
    insert(&pqueue, 2, 1.0)

    assert pqueue.size == 4
    A_ref = [2, 1, 3, 0]
    n_ref = [3, 1, 0, 2]
    key_ref = [2.0, 3.0, 1.0, 4.0]
    for i in range(4):
        assert pqueue.A[i] == A_ref[i]
        assert pqueue.Elements[i].node_idx == n_ref[i]
        assert pqueue.Elements[i].state == LABELED
        assert pqueue.Elements[i].key == key_ref[i]

    decrease_key(&pqueue, 3, 0.0)

    assert pqueue.size == 4
    A_ref = [3, 1, 2, 0]
    n_ref = [3, 1, 2, 0]
    key_ref = [2.0, 3.0, 1.0, 0.0]
    for i in range(4):
        assert pqueue.A[i] == A_ref[i]
        assert pqueue.Elements[i].node_idx == n_ref[i]
        assert pqueue.Elements[i].state == LABELED
        assert pqueue.Elements[i].key == key_ref[i]

    decrease_key(&pqueue, 1, -1.0)

    assert pqueue.size == 4
    A_ref = [1, 3, 2, 0]
    n_ref = [3, 0, 2, 1]
    key_ref = [2.0, -1.0, 1.0, 0.0]
    for i in range(4):
        assert pqueue.A[i] == A_ref[i]
        assert pqueue.Elements[i].node_idx == n_ref[i]
        assert pqueue.Elements[i].state == LABELED
        assert pqueue.Elements[i].key == key_ref[i]

    free_pqueue(&pqueue)


cdef void heapsort(DTYPE_t[::1] values_in, DTYPE_t[::1] values_out) nogil:
    """ 
    Heap sort by inerting all the values into the priority queue,
    and extracting them.
    """

    cdef:
        size_t i, l = <size_t>values_in.shape[0]
        PriorityQueue pqueue
    
    init_pqueue(&pqueue, l, l)
    for i in range(l):
        insert(&pqueue, i, values_in[i])
    for i in range(l):
        values_out[i] = pqueue.Elements[extract_min(&pqueue)].key
    free_pqueue(&pqueue)


cpdef sort_01(int n, random_seed=124):
    """ 
    Create an array of random number, sort it with the heapsort function 
    and with the numpy default sort function, compare the results.
    """
    
    cdef PriorityQueue pqueue

    rng = np.random.default_rng(random_seed)
    values_in = rng.random(size=n)
    values_out = np.empty_like(values_in, dtype=DTYPE_PY)
    heapsort(values_in, values_out)
    values_in_sorted = np.sort(values_in)
    np.testing.assert_array_equal(values_in_sorted, values_out)



""" 
Path-related methods.
"""

import warnings

import numpy as np
import pandas as pd

from edsger.commons import (
    A_VERY_SMALL_TIME_INTERVAL_PY,
    DTYPE_INF_PY,
    DTYPE_PY,
    INF_FREQ_PY,
    MIN_FREQ_PY,
)
from edsger.dijkstra import (
    compute_sssp,
    compute_sssp_w_path,
    compute_stsp,
    compute_stsp_w_path,
)
from edsger.spiess_florian import compute_SF_in
from edsger.star import (
    convert_graph_to_csc_float64,
    convert_graph_to_csc_uint32,
    convert_graph_to_csr_float64,
    convert_graph_to_csr_uint32,
)
from edsger.path_tracking import compute_path


class Dijkstra:
    """
    Dijkstra's algorithm for finding the shortest paths between nodes in directed graphs with
    positive edge weights.

    Parameters:
    -----------
    edges: pandas.DataFrame
        DataFrame containing the edges of the graph. It should have three columns: 'tail', 'head',
        and 'weight'. The 'tail' column should contain the IDs of the starting nodes, the 'head'
        column should contain the IDs of the ending nodes, and the 'weight' column should contain
        the (positive) weights of the edges.

    tail: str, optional (default='tail')
        The name of the column in the DataFrame that contains the IDs of the starting nodes.

    head: str, optional (default='head')
        The name of the column in the DataFrame that contains the IDs of the ending nodes.

    weight: str, optional (default='weight')
        The name of the column in the DataFrame that contains the (positive) weights of the edges.

    orientation: str, optional (default='out')
        The orientation of Dijkstra's algorithm. It can be either 'out' for single source shortest
        paths or 'in' for single target shortest path.

    check_edges: bool, optional (default=False)
        Whether to check if the edges DataFrame is well-formed. If set to True, the edges DataFrame
        will be checked for missing values and invalid data types.

    permute: bool, optional (default=False)
        Whether to permute the IDs of the nodes. If set to True, the node IDs will be reindexed to
        start from 0 and be contiguous.

    Attributes:
    -----------
    _edges: pandas.DataFrame
        DataFrame containing the edges of the graph.

    _n_edges: int
        The number of edges in the graph.

    _permute: bool
        Whether to permute the IDs of the nodes.

    _vertices: pandas.DataFrame or None
        DataFrame containing the old and new IDs of the nodes if the IDs have been permuted.

    _n_vertices: int
        The number of nodes in the graph (after permutation, if _permute is True).

    __n_vertices_init: int
        The number of nodes in the original graph (not permuted).

    _orientation: str
        The orientation of Dijkstra's algorithm.

    __indices: numpy.ndarray
        1D array containing the indices of the indices of the forward or reverse star of the graph
        in compressed format.

    __indptr: numpy.ndarray
        1D array containing the indices of the pointer of the forward or reverse star of the graph
        in compressed format.

    __edge_weights: numpy.ndarray
        1D array containing the weights of the edges in the graph.

    _path_links: numpy.ndarray
        predecessors or successors node index if the path tracking is activated.

    Methods:
    --------
    _check_edges(edges, tail, head, weight)
        Checks if the edges DataFrame is well-formed. If not, raises an appropriate error.

    _permute_graph(tail, head)
        Permute the IDs of the nodes to start from 0 and be contiguous. Returns a DataFrame with
        the permuted IDs.

    _check_orientation(orientation):
        Checks the orientation attribute.

    """

    def __init__(
        self,
        edges,
        tail="tail",
        head="head",
        weight="weight",
        orientation="out",
        check_edges=False,
        permute=False,
    ):
        # load the edges
        if check_edges:
            self._check_edges(edges, tail, head, weight)
        self._edges = edges[[tail, head, weight]].copy(deep=True)
        self._n_edges = len(self._edges)

        # reindex the vertices
        self._permute = permute
        if self._permute:
            self.__n_vertices_init = self._edges[[tail, head]].max(axis=0).max() + 1
            self._vertices = self._permute_graph(tail, head)
            self._n_vertices = len(self._vertices)
        else:
            self._vertices = None
            self._n_vertices = self._edges[[tail, head]].max(axis=0).max() + 1
            self.__n_vertices_init = self._n_vertices

        # convert to CSR/CSC
        self._check_orientation(orientation)
        self._orientation = orientation
        if self._orientation == "out":
            fs_indptr, fs_indices, fs_data = convert_graph_to_csr_float64(
                self._edges, tail, head, weight, self._n_vertices
            )
            self.__indices = fs_indices.astype(np.uint32)
            self.__indptr = fs_indptr.astype(np.uint32)
            self.__edge_weights = fs_data.astype(DTYPE_PY)
        else:
            rs_indptr, rs_indices, rs_data = convert_graph_to_csc_float64(
                self._edges, tail, head, weight, self._n_vertices
            )
            self.__indices = rs_indices.astype(np.uint32)
            self.__indptr = rs_indptr.astype(np.uint32)
            self.__edge_weights = rs_data.astype(DTYPE_PY)

        self._path_links = None

    @property
    def edges(self):
        return self._edges

    @property
    def vertices(self):
        return self._vertices

    @property
    def n_edges(self):
        return self._n_edges

    @property
    def n_vertices(self):
        return self._n_vertices

    @property
    def orientation(self):
        return self._orientation

    @property
    def permute(self):
        return self._permute

    @property
    def permute(self):
        return self._path_links

    def _check_edges(self, edges, tail, head, weight):
        """Checks if the edges DataFrame is well-formed. If not, raises an appropriate error."""
        if type(edges) != pd.core.frame.DataFrame:
            raise TypeError("edges should be a pandas DataFrame")

        if tail not in edges:
            raise KeyError(
                f"edge tail column '{tail}' not found in graph edges dataframe"
            )

        if head not in edges:
            raise KeyError(
                f"edge head column '{head}' not found in graph edges dataframe"
            )

        if weight not in edges:
            raise KeyError(
                f"edge weight column '{weight}' not found in graph edges dataframe"
            )

        if edges[[tail, head, weight]].isna().any().any():
            raise ValueError(
                " ".join(
                    [
                        f"edges[[{tail}, {head}, {weight}]] ",
                        "should not have any missing value",
                    ]
                )
            )

        for col in [tail, head]:
            if not pd.api.types.is_integer_dtype(edges[col].dtype):
                raise TypeError(f"edges['{col}'] should be of integer type")

        if not pd.api.types.is_numeric_dtype(edges[weight].dtype):
            raise TypeError(f"edges['{weight}'] should be of numeric type")

        if edges[weight].min() < 0.0:
            raise ValueError(f"edges['{weight}'] should be nonnegative")

        if not np.isfinite(edges[weight]).all():
            raise ValueError(f"edges['{weight}'] should be finite")

    def _permute_graph(self, tail, head):
        """Permute the IDs of the nodes to start from 0 and be contiguous.
        Returns a DataFrame with the permuted IDs."""

        vertices = pd.DataFrame(
            data={
                "vert_idx": np.union1d(
                    self._edges[tail].values, self._edges[head].values
                )
            }
        )
        vertices["vert_idx_new"] = vertices.index
        vertices.index.name = "index"

        self._edges = pd.merge(
            self._edges,
            vertices[["vert_idx", "vert_idx_new"]],
            left_on=tail,
            right_on="vert_idx",
            how="left",
        )
        self._edges.drop([tail, "vert_idx"], axis=1, inplace=True)
        self._edges.rename(columns={"vert_idx_new": tail}, inplace=True)

        self._edges = pd.merge(
            self._edges,
            vertices[["vert_idx", "vert_idx_new"]],
            left_on=head,
            right_on="vert_idx",
            how="left",
        )
        self._edges.drop([head, "vert_idx"], axis=1, inplace=True)
        self._edges.rename(columns={"vert_idx_new": head}, inplace=True)

        vertices.rename(columns={"vert_idx": "vert_idx_old"}, inplace=True)
        vertices.reset_index(drop=True, inplace=True)
        vertices.sort_values(by="vert_idx_new", inplace=True)

        vertices.index.name = "index"
        self._edges.index.name = "index"

        return vertices

    def _check_orientation(self, orientation):
        """Checks the orientation attribute."""
        if orientation not in ["in", "out"]:
            raise ValueError(f"orientation should be either 'in' on 'out'")

    def run(
        self,
        vertex_idx,
        path_tracking=False,
        return_inf=True,
        return_Series=False,
        heap_length_ratio=1.0,
    ):
        """
        Runs shortest path algorithm between a given vertex and all other vertices in the graph.

        Parameters
        ----------
        vertex_idx : int
            The index of the source/target vertex.
        path_tracking : bool, optional (default=False)
            Whether to track the shortest path(s) from the source vertex to all other vertices in
            the graph.
        return_inf : bool, optional (default=True)
            Whether to return path length(s) as infinity (np.inf) when no path exists.
        return_Series : bool, optional (default=False)
            Whether to return a Pandas Series object indexed by vertex indices with path length(s)
            as values.
        heap_length_ratio : float, optional (default=1.0)
            The heap length as a fraction of the number of vertices. Must be in the range (0, 1].

        Returns
        -------
        path_length_values : array_like or Pandas Series
            If `return_Series=False`, a 1D Numpy array of shape (n_vertices,) with the shortest
            path length from the source vertex to each vertex in the graph (`orientation="out"`), or
            from each vertex to the target vertex (`orientation="in"`). If `return_Series=True`, a
            Pandas Series object with the same data
            and the vertex indices as index.

        """
        # validate the input arguments
        if not isinstance(vertex_idx, int):
            raise TypeError(f"argument 'vertex_idx=f{vertex_idx}' must be of int type")
        if vertex_idx < 0:
            raise ValueError(f"argument 'vertex_idx={vertex_idx}' must be positive")
        if self._permute:
            if vertex_idx not in self._vertices.vert_idx_old.values:
                raise ValueError(f"vertex {vertex_idx} not found in graph")
            vertex_new = self._vertices.loc[
                self._vertices.vert_idx_old == vertex_idx, "vert_idx_new"
            ].iloc[0]
        else:
            if vertex_idx >= self._n_vertices:
                raise ValueError(f"vertex {vertex_idx} not found in graph")
            vertex_new = vertex_idx
        if not isinstance(path_tracking, bool):
            raise TypeError(
                f"argument 'path_tracking=f{path_tracking}' must be of bool type"
            )
        if not isinstance(return_inf, bool):
            raise TypeError(f"argument 'return_inf=f{return_inf}' must be of bool type")
        if not isinstance(return_Series, bool):
            raise TypeError(
                f"argument 'return_Series=f{return_Series}' must be of bool type"
            )
        if not isinstance(heap_length_ratio, float):
            raise TypeError(
                f"argument 'heap_length_ratio=f{heap_length_ratio}' must be of float type"
            )
        if heap_length_ratio > 1.0:
            heap_length_ratio = 1.0
        if heap_length_ratio <= 0.0:
            raise ValueError(
                f"argument 'heap_length_ratio={heap_length_ratio}' must be strictly positive "
            )
        heap_length = int(np.rint(heap_length_ratio * self._n_vertices))

        # compute path length
        if not path_tracking:
            self._path_links = None
            if self._orientation == "in":
                path_length_values = compute_stsp(
                    self.__indptr,
                    self.__indices,
                    self.__edge_weights,
                    vertex_new,
                    self._n_vertices,
                    heap_length,
                )
            else:
                path_length_values = compute_sssp(
                    self.__indptr,
                    self.__indices,
                    self.__edge_weights,
                    vertex_new,
                    self._n_vertices,
                    heap_length,
                )
        else:
            self._path_links = np.arange(0, self._n_vertices, dtype=np.uint32)
            if self._orientation == "in":
                path_length_values = compute_stsp_w_path(
                    self.__indptr,
                    self.__indices,
                    self.__edge_weights,
                    self._path_links,
                    vertex_new,
                    self._n_vertices,
                    heap_length,
                )
            else:
                path_length_values = compute_sssp_w_path(
                    self.__indptr,
                    self.__indices,
                    self.__edge_weights,
                    self._path_links,
                    vertex_new,
                    self._n_vertices,
                    heap_length,
                )

            if self._permute:
                # permute back the path vertex indices
                path_df = pd.DataFrame(
                    data={
                        "vertex_idx": np.arange(self._n_vertices),
                        "associated_idx": self._path_links,
                    }
                )
                path_df = pd.merge(
                    path_df,
                    self._vertices,
                    left_on="vertex_idx",
                    right_on="vert_idx_new",
                    how="left",
                )
                path_df.drop(["vertex_idx", "vert_idx_new"], axis=1, inplace=True)
                path_df.rename(columns={"vert_idx_old": "vertex_idx"}, inplace=True)
                path_df = pd.merge(
                    path_df,
                    self._vertices,
                    left_on="associated_idx",
                    right_on="vert_idx_new",
                    how="left",
                )
                path_df.drop(["associated_idx", "vert_idx_new"], axis=1, inplace=True)
                path_df.rename(columns={"vert_idx_old": "associated_idx"}, inplace=True)

                if return_Series:
                    path_df.set_index("vertex_idx", inplace=True)
                    self._path_links = path_df.associated_idx.astype(np.uint32)
                else:
                    self._path_links = np.arange(
                        self.__n_vertices_init, dtype=np.uint32
                    )
                    self._path_links[
                        path_df.vertex_idx.values
                    ] = path_df.associated_idx.values

        # deal with infinity
        if return_inf:
            path_length_values = np.where(
                path_length_values == DTYPE_INF_PY, np.inf, path_length_values
            )

        # reorder path lengths
        if return_Series:
            if self._permute:
                self._vertices["path_length"] = path_length_values
                path_lengths_df = self._vertices[
                    ["vert_idx_old", "path_length"]
                ].sort_values(by="vert_idx_old")
                path_lengths_df.set_index("vert_idx_old", drop=True, inplace=True)
                path_lengths_df.index.name = "vertex_idx"
                path_lengths_series = path_lengths_df.path_length
            else:
                path_lengths_series = pd.Series(path_length_values)
                path_lengths_series.index.name = "vertex_idx"
                path_lengths_series.name = "path_length"

            return path_lengths_series

        else:
            if self._permute:
                self._vertices["path_length"] = path_length_values
                if return_inf:
                    path_length_values = np.inf * np.ones(self.__n_vertices_init)
                else:
                    path_length_values = DTYPE_INF_PY * np.ones(self.__n_vertices_init)
                path_length_values[
                    self._vertices.vert_idx_old.values
                ] = self._vertices.path_length.values

            return path_length_values

    def get_path(self, vertex_idx):
        if self._path_links is None:
            warnings.warn(
                "Current Dijkstra instance has not path attribute : \
                make sure path_tracking is set to True, and run the \
                shortest path algorithm",
                UserWarning,
            )
        else:
            if isinstance(self._path_links, pd.Series):
                path_vertices = compute_path(self._path_links.values, vertex_idx)
            else:
                path_vertices = compute_path(self._path_links, vertex_idx)
            return path_vertices


class HyperpathGenerating:
    def __init__(
        self,
        edges,
        tail="tail",
        head="head",
        trav_time="trav_time",
        freq="freq",
        check_edges=False,
        orientation="in",
    ):
        # load the edges
        if check_edges:
            self._check_edges(edges, tail, head, trav_time, freq)
        self._edges = edges[[tail, head, trav_time, freq]].copy(deep=True)
        self.edge_count = len(self._edges)

        # remove inf values if any, and values close to zero
        self._edges[trav_time] = np.where(
            self._edges[trav_time] > DTYPE_INF_PY, DTYPE_INF_PY, self._edges[trav_time]
        )
        self._edges[trav_time] = np.where(
            self._edges[trav_time] < A_VERY_SMALL_TIME_INTERVAL_PY,
            A_VERY_SMALL_TIME_INTERVAL_PY,
            self._edges[trav_time],
        )
        self._edges[freq] = np.where(
            self._edges[freq] > INF_FREQ_PY, INF_FREQ_PY, self._edges[freq]
        )
        self._edges[freq] = np.where(
            self._edges[freq] < MIN_FREQ_PY, MIN_FREQ_PY, self._edges[freq]
        )

        # create an edge index column
        self._edges = self._edges.reset_index(drop=True)
        data_col = "edge_idx"
        self._edges[data_col] = self._edges.index

        # convert to CSR/CSC format
        self.vertex_count = self._edges[[tail, head]].max().max() + 1
        assert orientation in ["out", "in"]
        self._orientation = orientation
        if self._orientation == "out":
            fs_indptr, _, fs_data = convert_graph_to_csr_uint32(
                self._edges, tail, head, data_col, self.vertex_count
            )
            self.__indptr = fs_indptr.astype(np.uint32)
            self._edge_idx = fs_data.astype(np.uint32)
        else:
            rs_indptr, _, rs_data = convert_graph_to_csc_uint32(
                self._edges, tail, head, data_col, self.vertex_count
            )
            self.__indptr = rs_indptr.astype(np.uint32)
            self._edge_idx = rs_data.astype(np.uint32)

        # edge attributes
        self._trav_time = self._edges[trav_time].values.astype(DTYPE_PY)
        self._freq = self._edges[freq].values.astype(DTYPE_PY)
        self._tail = self._edges[tail].values.astype(np.uint32)
        self._head = self._edges[head].values.astype(np.uint32)

    def run(self, origin, destination, volume, return_inf=False):
        # column storing the resulting edge volumes
        self._edges["volume"] = 0.0
        self.u_i_vec = None

        # vertex least travel time
        u_i_vec = DTYPE_INF_PY * np.ones(self.vertex_count, dtype=DTYPE_PY)

        # input check
        if type(volume) is not list:
            volume = [volume]
        if self._orientation == "out":
            self._check_vertex_idx(origin)
            if type(destination) is not list:
                destination = [destination]
            assert len(destination) == len(volume)
            for i, item in enumerate(destination):
                self._check_vertex_idx(item)
                self._check_volume(volume[i])
            demand_indices = np.array(destination, dtype=np.uint32)
        elif self._orientation == "in":
            if type(origin) is not list:
                origin = [origin]
            assert len(origin) == len(volume)
            for i, item in enumerate(origin):
                self._check_vertex_idx(item)
                self._check_volume(volume[i])
            self._check_vertex_idx(destination)
            demand_indices = np.array(origin, dtype=np.uint32)
        assert isinstance(return_inf, bool)

        demand_values = np.array(volume, dtype=DTYPE_PY)

        if self._orientation == "out":
            raise NotImplementedError(
                "one-to-many Spiess & Florian's algorithm not implemented yet"
            )
        elif self._orientation == "in":
            compute_SF_in(
                self.__indptr,
                self._edge_idx,
                self._trav_time,
                self._freq,
                self._tail,
                self._head,
                demand_indices,  # source vertex indices
                demand_values,
                self._edges["volume"].values,
                u_i_vec,
                self.vertex_count,
                destination,
            )
            self.u_i_vec = u_i_vec

    def _check_vertex_idx(self, idx):
        assert isinstance(idx, int)
        assert idx >= 0
        assert idx < self.vertex_count

    def _check_volume(self, v):
        assert isinstance(v, float)
        assert v >= 0.0

    def _check_edges(self, edges, tail, head, trav_time, freq):
        if type(edges) != pd.core.frame.DataFrame:
            raise TypeError("edges should be a pandas DataFrame")

        for col in [tail, head, trav_time, freq]:
            if col not in edges:
                raise KeyError(
                    f"edge column '{col}' not found in graph edges dataframe"
                )

        if edges[[tail, head, trav_time, freq]].isna().any().any():
            raise ValueError(
                " ".join(
                    [
                        f"edges[[{tail}, {head}, {trav_time}, {freq}]] ",
                        "should not have any missing value",
                    ]
                )
            )

        for col in [tail, head]:
            if not pd.api.types.is_integer_dtype(edges[col].dtype):
                raise TypeError(f"column '{col}' should be of integer type")

        for col in [trav_time, freq]:
            if not pd.api.types.is_numeric_dtype(edges[col].dtype):
                raise TypeError(f"column '{col}' should be of numeric type")

            if edges[col].min() < 0.0:
                raise ValueError(f"column '{col}' should be nonnegative")



""" 
An implementation of Dijkstra's algorithm.
    
author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

cimport numpy as cnp

from edsger.commons cimport (
    DTYPE_INF, UNLABELED, SCANNED, DTYPE_t, ElementState)
cimport edsger.pq_4ary_dec_0b as pq  # priority queue


cpdef cnp.ndarray compute_sssp(
    cnp.uint32_t[::1] csr_indptr,
    cnp.uint32_t[::1] csr_indices,
    DTYPE_t[::1] csr_data,
    int source_vert_idx,
    int vertex_count,
    int heap_length):
    """ 
    Compute single-source shortest path (from one vertex to all vertices)
    using the pq_4ary_dec_0b priority queue.

    Does not return predecessors.

    input
    =====
    * cnp.uint32_t[::1] csr_indices : indices in the CSR format
    * cnp.uint32_t[::1] csr_indptr : pointers in the CSR format
    * DTYPE_t[::1] csr_data :  data (edge weights) in the CSR format
    * int source_vert_idx : source vertex index
    * int vertex_count : vertex count
    * int heap_length : heap length

    output
    ======
    * cnp.ndarray : shortest path length for each vertex
    """

    cdef:
        size_t tail_vert_idx, head_vert_idx, idx
        DTYPE_t tail_vert_val, head_vert_val
        pq.PriorityQueue pqueue
        ElementState vert_state
        size_t source = <size_t>source_vert_idx

    with nogil:

        # initialization of the heap elements 
        # all nodes have INFINITY key and UNLABELED state
        pq.init_pqueue(&pqueue, <size_t>heap_length, <size_t>vertex_count)

        # the key is set to zero for the source vertex,
        # which is inserted into the heap
        pq.insert(&pqueue, source, 0.0)

        # main loop
        while pqueue.size > 0:
            tail_vert_idx = pq.extract_min(&pqueue)
            tail_vert_val = pqueue.Elements[tail_vert_idx].key

            # loop on outgoing edges
            for idx in range(<size_t>csr_indptr[tail_vert_idx], 
                <size_t>csr_indptr[tail_vert_idx + 1]):

                head_vert_idx = <size_t>csr_indices[idx]
                vert_state = pqueue.Elements[head_vert_idx].state
                if vert_state != SCANNED:
                    head_vert_val = tail_vert_val + csr_data[idx]
                    if vert_state == UNLABELED:
                        pq.insert(&pqueue, head_vert_idx, head_vert_val)
                    elif pqueue.Elements[head_vert_idx].key > head_vert_val:
                        pq.decrease_key(&pqueue, head_vert_idx, head_vert_val)

    # copy the results into a numpy array
    path_lengths = pq.copy_keys_to_numpy(&pqueue, <size_t>vertex_count)

    # cleanup
    pq.free_pqueue(&pqueue)  

    return path_lengths


cpdef cnp.ndarray compute_sssp_w_path(
    cnp.uint32_t[::1] csr_indptr,
    cnp.uint32_t[::1] csr_indices,
    DTYPE_t[::1] csr_data,
    cnp.uint32_t[::1] predecessor,
    int source_vert_idx,
    int vertex_count,
    int heap_length):
    """ 
    Compute single-source shortest path (from one vertex to all vertices)
    using the pq_4ary_dec_0b priority queue.

    Compute predecessors.

    input
    =====
    * cnp.uint32_t[::1] csr_indices : indices in the CSR format
    * cnp.uint32_t[::1] csr_indptr : pointers in the CSR format
    * DTYPE_t[::1] csr_data :  data (edge weights) in the CSR format
    * cnp.uint32_t[::1] predecessor :  array of indices, one for each vertex of 
        the graph. Each vertex' entry contains the index of its predecessor in 
        a path from the source, through the graph. 
    * int source_vert_idx : source vertex index
    * int vertex_count : vertex count
    * int heap_length : heap length

    output
    ======
    * cnp.ndarray : shortest path length for each vertex
    """

    cdef:
        size_t tail_vert_idx, head_vert_idx, idx
        DTYPE_t tail_vert_val, head_vert_val
        pq.PriorityQueue pqueue
        ElementState vert_state
        size_t source = <size_t>source_vert_idx

    with nogil:

        # initialization of the heap elements 
        # all nodes have INFINITY key and UNLABELED state
        pq.init_pqueue(&pqueue, <size_t>heap_length, <size_t>vertex_count)

        # the key is set to zero for the source vertex,
        # which is inserted into the heap
        pq.insert(&pqueue, source, 0.0)

        # main loop
        while pqueue.size > 0:
            tail_vert_idx = pq.extract_min(&pqueue)
            tail_vert_val = pqueue.Elements[tail_vert_idx].key

            # loop on outgoing edges
            for idx in range(<size_t>csr_indptr[tail_vert_idx], 
                <size_t>csr_indptr[tail_vert_idx + 1]):

                head_vert_idx = <size_t>csr_indices[idx]
                vert_state = pqueue.Elements[head_vert_idx].state
                if vert_state != SCANNED:
                    head_vert_val = tail_vert_val + csr_data[idx]
                    if vert_state == UNLABELED:
                        pq.insert(&pqueue, head_vert_idx, head_vert_val)
                        predecessor[head_vert_idx] = tail_vert_idx
                    elif pqueue.Elements[head_vert_idx].key > head_vert_val:
                        pq.decrease_key(&pqueue, head_vert_idx, head_vert_val)
                        predecessor[head_vert_idx] = tail_vert_idx

    # copy the results into a numpy array
    path_lengths = pq.copy_keys_to_numpy(&pqueue, <size_t>vertex_count)

    # cleanup
    pq.free_pqueue(&pqueue)  

    return path_lengths


cpdef cnp.ndarray compute_stsp(
    cnp.uint32_t[::1] csc_indptr,
    cnp.uint32_t[::1] csc_indices,
    DTYPE_t[::1] csc_data,
    int target_vert_idx,
    int vertex_count,
    int heap_length):
    """ 
    Compute single-target shortest path (from all vertices to one vertex)
    using the pq_4ary_dec_0b priority queue.

    Does not return successors.

    input
    =====
    * cnp.uint32_t[::1] csc_indices : indices in the CSC format
    * cnp.uint32_t[::1] csc_indptr : pointers in the CSC format
    * DTYPE_t[::1] csc_data :  data (edge weights) in the CSC format
    * int target_vert_idx : source vertex index
    * int vertex_count : vertex count
    * int heap_length : heap length

    output
    ======
    * cnp.ndarray : shortest path length for each vertex
    """

    cdef:
        size_t tail_vert_idx, head_vert_idx, idx
        DTYPE_t tail_vert_val, head_vert_val
        pq.PriorityQueue pqueue
        ElementState vert_state
        size_t target = <size_t>target_vert_idx

    with nogil:

        # initialization of the heap elements 
        # all nodes have INFINITY key and UNLABELED state
        pq.init_pqueue(&pqueue, <size_t>heap_length, <size_t>vertex_count)

        # the key is set to zero for the target vertex,
        # which is inserted into the heap
        pq.insert(&pqueue, target, 0.0)

        # main loop
        while pqueue.size > 0:
            head_vert_idx = pq.extract_min(&pqueue)
            head_vert_val = pqueue.Elements[head_vert_idx].key

            # loop on incoming edges
            for idx in range(<size_t>csc_indptr[head_vert_idx], 
                <size_t>csc_indptr[head_vert_idx + 1]):

                tail_vert_idx = <size_t>csc_indices[idx]
                vert_state = pqueue.Elements[tail_vert_idx].state
                if vert_state != SCANNED:
                    tail_vert_val = head_vert_val + csc_data[idx]
                    if vert_state == UNLABELED:
                        pq.insert(&pqueue, tail_vert_idx, tail_vert_val)
                    elif pqueue.Elements[tail_vert_idx].key > tail_vert_val:
                        pq.decrease_key(&pqueue, tail_vert_idx, tail_vert_val)

    # copy the results into a numpy array
    path_lengths = pq.copy_keys_to_numpy(&pqueue, <size_t>vertex_count)

    # cleanup
    pq.free_pqueue(&pqueue)  

    return path_lengths


cpdef cnp.ndarray compute_stsp_w_path(
    cnp.uint32_t[::1] csc_indptr,
    cnp.uint32_t[::1] csc_indices,
    DTYPE_t[::1] csc_data,
    cnp.uint32_t[::1] successor,
    int target_vert_idx,
    int vertex_count,
    int heap_length):
    """ 
    Compute single-target shortest path (from all vertices to one vertex)
    using the pq_4ary_dec_0b priority queue.

    Compute successors.

    input
    =====
    * cnp.uint32_t[::1] csc_indices : indices in the CSC format
    * cnp.uint32_t[::1] csc_indptr : pointers in the CSC format
    * DTYPE_t[::1] csc_data :  data (edge weights) in the CSC format
    * cnp.uint32_t[::1] successor :  array of indices, one for each vertex of 
        the graph. Each vertex' entry contains the index of its successor in 
        a path to the target, through the graph. 
    * int target_vert_idx : source vertex index
    * int vertex_count : vertex count
    * int heap_length : heap length

    output
    ======
    * cnp.ndarray : shortest path length for each vertex
    """

    cdef:
        size_t tail_vert_idx, head_vert_idx, idx
        DTYPE_t tail_vert_val, head_vert_val
        pq.PriorityQueue pqueue
        ElementState vert_state
        size_t target = <size_t>target_vert_idx

    with nogil:

        # initialization of the heap elements 
        # all nodes have INFINITY key and UNLABELED state
        pq.init_pqueue(&pqueue, <size_t>heap_length, <size_t>vertex_count)

        # the key is set to zero for the target vertex,
        # which is inserted into the heap
        pq.insert(&pqueue, target, 0.0)

        # main loop
        while pqueue.size > 0:
            head_vert_idx = pq.extract_min(&pqueue)
            head_vert_val = pqueue.Elements[head_vert_idx].key

            # loop on incoming edges
            for idx in range(<size_t>csc_indptr[head_vert_idx], 
                <size_t>csc_indptr[head_vert_idx + 1]):

                tail_vert_idx = <size_t>csc_indices[idx]
                vert_state = pqueue.Elements[tail_vert_idx].state
                if vert_state != SCANNED:
                    tail_vert_val = head_vert_val + csc_data[idx]
                    if vert_state == UNLABELED:
                        pq.insert(&pqueue, tail_vert_idx, tail_vert_val)
                        successor[tail_vert_idx] = head_vert_idx
                    elif pqueue.Elements[tail_vert_idx].key > tail_vert_val:
                        pq.decrease_key(&pqueue, tail_vert_idx, tail_vert_val)
                        successor[tail_vert_idx] = head_vert_idx

    # copy the results into a numpy array
    path_lengths = pq.copy_keys_to_numpy(&pqueue, <size_t>vertex_count)

    # cleanup
    pq.free_pqueue(&pqueue)  

    return path_lengths


# ============================================================================ #
# tests                                                                        #
# ============================================================================ #

from edsger.commons import DTYPE_PY
import numpy as np


cdef generate_single_edge_network_csr():
    """  
    Generate a single edge network in CSR format.

    This network has 1 edge and 2 vertices.
    """

    csr_indptr = np.array([0, 1, 1], dtype=np.uint32)
    csr_indices = np.array([1], dtype=np.uint32)
    csr_data = np.array([1.], dtype=DTYPE_PY)

    return csr_indptr, csr_indices, csr_data


cdef generate_single_edge_network_csc():
    """  
    Generate a single edge network in CSC format.

    This network has 1 edge and 2 vertices.
    """

    csc_indptr = np.array([0, 0, 1], dtype=np.uint32)
    csc_indices = np.array([0], dtype=np.uint32)
    csc_data = np.array([1.], dtype=DTYPE_PY)

    return csc_indptr, csc_indices, csc_data


cdef generate_braess_network_csr():
    """ 
    Generate a Braess-like network in CSR format.

    This network hs 5 edges and 4 vertices.
    """

    csr_indptr = np.array([0, 2, 4, 5, 5], dtype=np.uint32)
    csr_indices = np.array([1, 2, 2, 3, 3], dtype=np.uint32)
    csr_data = np.array([1., 2., 0., 2., 1.], dtype=DTYPE_PY)

    return csr_indptr, csr_indices, csr_data


cdef generate_braess_network_csc():
    """ 
    Generate a Braess-like network in CSC format.

    This network hs 5 edges and 4 vertices.
    """

    csc_indptr = np.array([0, 0, 1, 3, 5], dtype=np.uint32)
    csc_indices = np.array([0, 0, 1, 1, 2], dtype=np.uint32)
    csc_data = np.array([1., 2., 0., 2., 1.], dtype=DTYPE_PY)

    return csc_indptr, csc_indices, csc_data


cpdef compute_sssp_01():
    """ 
    Compute SSSP with the compute_sssp_pq_bd0 routine on a single edge 
    network.
    """

    csr_indptr, csr_indices, csr_data = generate_single_edge_network_csr()

    # from vertex 0
    path_lengths = compute_sssp(csr_indptr, csr_indices, csr_data, 0, 2, 2)
    path_lengths_ref = np.array([0., 1.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 1
    path_lengths = compute_sssp(csr_indptr, csr_indices, csr_data, 1, 2, 2)
    path_lengths_ref = np.array([DTYPE_INF, 0.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)


cpdef compute_stsp_01():
    """ 
    Compute TSSP with the compute_stsp_pq_bd0 routine on a single edge 
    network.
    """

    csc_indptr, csc_indices, csc_data = generate_single_edge_network_csc()

    # from vertex 0
    path_lengths = compute_stsp(csc_indptr, csc_indices, csc_data, 0, 2, 2)
    path_lengths_ref = np.array([0., DTYPE_INF], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 1
    path_lengths = compute_stsp(csc_indptr, csc_indices, csc_data, 1, 2, 2)
    path_lengths_ref = np.array([1., 0.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)


cpdef compute_sssp_02():
    """ 
    Compute SSSP with the compute_sssp_pq_bd0 routine on Braess-like 
    network.
    """

    csr_indptr, csr_indices, csr_data = generate_braess_network_csr()

    # from vertex 0
    path_lengths = compute_sssp(csr_indptr, csr_indices, csr_data, 0, 4, 4)
    path_lengths_ref = np.array([0., 1., 1., 2.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 1
    path_lengths = compute_sssp(csr_indptr, csr_indices, csr_data, 1, 4, 4)
    path_lengths_ref = np.array([DTYPE_INF, 0., 0., 1.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 2
    path_lengths = compute_sssp(csr_indptr, csr_indices, csr_data, 2, 4, 4)
    path_lengths_ref = np.array([DTYPE_INF, DTYPE_INF, 0., 1.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 3
    path_lengths = compute_sssp(csr_indptr, csr_indices, csr_data, 3, 4, 4)
    path_lengths_ref = np.array([DTYPE_INF, DTYPE_INF, DTYPE_INF, 0.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)


cpdef compute_stsp_02():
    """ 
    Compute STSP with the compute_stsp_pq_bd0 routine on Braess-like 
    network.
    """

    csc_indptr, csc_indices, csc_data = generate_braess_network_csc()

    # from vertex 0
    path_lengths = compute_stsp(csc_indptr, csc_indices, csc_data, 0, 4, 4)
    path_lengths_ref = np.array([0., DTYPE_INF, DTYPE_INF, DTYPE_INF], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 1
    path_lengths = compute_stsp(csc_indptr, csc_indices, csc_data, 1, 4, 4)
    path_lengths_ref = np.array([1., 0., DTYPE_INF, DTYPE_INF], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 2
    path_lengths = compute_stsp(csc_indptr, csc_indices, csc_data, 2, 4, 4)
    path_lengths_ref = np.array([1., 0., 0., DTYPE_INF], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)

    # from vertex 3
    path_lengths = compute_stsp(csc_indptr, csc_indices, csc_data, 3, 4, 4)
    path_lengths_ref = np.array([2., 1.0, 1., 0.], dtype=DTYPE_PY)
    assert np.allclose(path_lengths_ref, path_lengths)


# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html


# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../../src/edsger/"))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Edsger"
copyright = "2023, Architecture & Performance"
author = "Francois Pacull"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

templates_path = ["_templates"]
exclude_patterns = ["*.pyx"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
# html_theme = "furo"
html_static_path = ["_static"]

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx_design",
]



"""
Forward and reverse star representations of networks.
"""

import numpy as np
cimport numpy as cnp


cpdef convert_graph_to_csr_uint32(edges, tail, head, data, vertex_count):
    """
    Convert an edge dataframe in COO format into CSR format.

    The data vector is of uint32 type.

    Parameters
    ----------
    edges : pandas.core.frame.DataFrame
        The edges dataframe.
    tail : str
        The column name in the edges dataframe for the tail vertex index.
    head : str
        The column name in the edges dataframe for the head vertex index.
    data : str
        The column name in the edges dataframe for the int edge attribute.
    vertex_count : int
        The vertex count in the given network edges.

    Returns
    -------
    tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """

    fs_indptr = np.zeros(
        vertex_count + 1, dtype=np.uint32
    )  # make sure it is filled with zeros
    edge_count = len(edges)
    fs_indices = np.empty(edge_count, dtype=np.uint32)
    fs_data = np.empty(edge_count, dtype=np.uint32)

    _coo_tocsr_uint32(
        edges[tail].values.astype(np.uint32),
        edges[head].values.astype(np.uint32),
        edges[data].values.astype(np.uint32),
        fs_indptr,
        fs_indices,
        fs_data,
    )

    return fs_indptr, fs_indices, fs_data


cpdef convert_graph_to_csc_uint32(edges, tail, head, data, vertex_count):
    """
    Convert an edge dataframe in COO format into CSC format.

    The data vector is of uint32 type.

    Parameters
    ----------
    edges : pandas.core.frame.DataFrame
        The edges dataframe.
    tail : str
        The column name in the edges dataframe for the tail vertex index.
    head : str
        The column name in the edges dataframe for the head vertex index.
    data : str
        The column name in the edges dataframe for the int edge attribute.
    vertex_count : int
        The vertex count in the given network edges.

    Returns
    -------
    tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """

    rs_indptr = np.zeros(
        vertex_count + 1, dtype=np.uint32
    )  # make sure it is filled with zeros
    edge_count = len(edges)
    rs_indices = np.empty(edge_count, dtype=np.uint32)
    rs_data = np.empty(edge_count, dtype=np.uint32)

    _coo_tocsc_uint32(
        edges[tail].values.astype(np.uint32),
        edges[head].values.astype(np.uint32),
        edges[data].values.astype(np.uint32),
        rs_indptr,
        rs_indices,
        rs_data,
    )

    return rs_indptr, rs_indices, rs_data


cpdef convert_graph_to_csr_float64(edges, tail, head, data, vertex_count):
    """
    Convert an edge dataframe in COO format into CSR format.

    The data vector is of float64 type.

    Parameters
    ----------
    edges : pandas.core.frame.DataFrame
        The edges dataframe.
    tail : str
        The column name in the edges dataframe for the tail vertex index.
    head : str
        The column name in the edges dataframe for the head vertex index.
    data : str
        The column name in the edges dataframe for the real edge attribute.
    vertex_count : int
        The vertex count in the given network edges.

    Returns
    -------
    tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """

    fs_indptr = np.zeros(
        vertex_count + 1, dtype=np.uint32
    )  # make sure it is filled with zeros
    edge_count = len(edges)
    fs_indices = np.empty(edge_count, dtype=np.uint32)
    fs_data = np.empty(edge_count, dtype=np.float64)

    _coo_tocsr_float64(
        edges[tail].values.astype(np.uint32),
        edges[head].values.astype(np.uint32),
        edges[data].values.astype(np.float64),
        fs_indptr,
        fs_indices,
        fs_data,
    )

    return fs_indptr, fs_indices, fs_data


cpdef convert_graph_to_csc_float64(edges, tail, head, data, vertex_count):
    """
    Convert an edge dataframe in COO format into CSC format.

    The data vector is of float64 type.

    Parameters
    ----------
    edges : pandas.core.frame.DataFrame
        The edges dataframe.
    tail : str
        The column name in the edges dataframe for the tail vertex index.
    head : str
        The column name in the edges dataframe for the head vertex index.
    data : str
        The column name in the edges dataframe for the real edge attribute.
    vertex_count : int
        The vertex count in the given network edges.

    Returns
    -------
    tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """

    rs_indptr = np.zeros(
        vertex_count + 1, dtype=np.uint32
    )  # make sure it is filled with zeros
    edge_count = len(edges)
    rs_indices = np.empty(edge_count, dtype=np.uint32)
    rs_data = np.empty(edge_count, dtype=np.float64)

    _coo_tocsc_float64(
        edges[tail].values.astype(np.uint32),
        edges[head].values.astype(np.uint32),
        edges[data].values.astype(np.float64),
        rs_indptr,
        rs_indices,
        rs_data,
    )

    return rs_indptr, rs_indices, rs_data


cdef void _coo_tocsr_uint32(
    cnp.uint32_t [::1] Ai,
    cnp.uint32_t [::1] Aj,
    cnp.uint32_t [::1] Ax,
    cnp.uint32_t [::1] Bp,
    cnp.uint32_t [::1] Bj,
    cnp.uint32_t [::1] Bx) nogil:    

    cdef:
        size_t i, row, dest
        size_t n_vert = <size_t>(Bp.shape[0] - 1)
        size_t n_edge = <size_t>Bj.shape[0]
        cnp.uint32_t temp, cumsum, last

    for i in range(n_edge):
        Bp[<size_t>Ai[i]] += 1

    cumsum = 0
    for i in range(n_vert):
        temp = Bp[i]
        Bp[i] = cumsum
        cumsum += temp
    Bp[n_vert] = <cnp.uint32_t>n_edge 

    for i in range(n_edge):
        row  = <size_t>Ai[i]
        dest = <size_t>Bp[row]
        Bj[dest] = Aj[i]
        Bx[dest] = Ax[i]
        Bp[row] += 1

    last = 0
    for i in range(n_vert + 1):
        temp = Bp[i]
        Bp[i] = last
        last = temp


cdef void _coo_tocsc_uint32(
    cnp.uint32_t [::1] Ai,
    cnp.uint32_t [::1] Aj,
    cnp.uint32_t [::1] Ax,   
    cnp.uint32_t [::1] Bp,
    cnp.uint32_t [::1] Bi,
    cnp.uint32_t [::1] Bx) nogil:

    cdef:
        size_t i, col, dest
        size_t n_vert = <size_t>(Bp.shape[0] - 1)
        size_t n_edge = <size_t>Bi.shape[0]
        cnp.uint32_t temp, cumsum, last

    for i in range(n_edge):
        Bp[<size_t>Aj[i]] += 1

    cumsum = 0
    for i in range(n_vert):
        temp = Bp[i]
        Bp[i] = cumsum
        cumsum += temp
    Bp[<size_t>n_vert] = <cnp.uint32_t>n_edge 

    for i in range(n_edge):
        col  = <size_t>Aj[i]
        dest = <size_t>Bp[col]
        Bi[dest] = Ai[i]
        Bx[dest] = Ax[i]
        Bp[col] += 1

    last = 0
    for i in range(n_vert + 1):
        temp = Bp[i]
        Bp[i] = last
        last = temp


cpdef void _coo_tocsr_float64(
    cnp.uint32_t  [::1] Ai,
    cnp.uint32_t  [::1] Aj,
    cnp.float64_t [::1] Ax,
    cnp.uint32_t  [::1] Bp,
    cnp.uint32_t  [::1] Bj,
    cnp.float64_t [::1] Bx) nogil:

    cdef:
        size_t i, row, dest
        size_t n_vert = <size_t>(Bp.shape[0] - 1)
        size_t n_edge = <size_t>Bj.shape[0]
        cnp.uint32_t temp, cumsum, last

    for i in range(n_edge):
        Bp[<size_t>Ai[i]] += 1

    cumsum = 0
    for i in range(n_vert):
        temp = Bp[i]
        Bp[i] = cumsum
        cumsum += temp
    Bp[n_vert] = <cnp.uint32_t>n_edge 

    for i in range(n_edge):
        row  = <size_t>Ai[i]
        dest = <size_t>Bp[row]
        Bj[dest] = Aj[i]
        Bx[dest] = Ax[i]
        Bp[row] += 1

    last = 0
    for i in range(n_vert + 1):
        temp = Bp[i]
        Bp[i] = last
        last = temp


cpdef void _coo_tocsc_float64(
    cnp.uint32_t  [::1] Ai,
    cnp.uint32_t  [::1] Aj,
    cnp.float64_t [::1] Ax,
    cnp.uint32_t  [::1] Bp,
    cnp.uint32_t  [::1] Bi,
    cnp.float64_t [::1] Bx) nogil:

    cdef:
        size_t i, col, dest
        size_t n_vert = <size_t>(Bp.shape[0] - 1)
        size_t n_edge = <size_t>Bi.shape[0]
        cnp.uint32_t temp, cumsum, last

    for i in range(n_edge):
        Bp[<size_t>Aj[i]] += 1

    cumsum = 0
    for i in range(n_vert):
        temp = Bp[i]
        Bp[i] = cumsum
        cumsum += temp
    Bp[<size_t>n_vert] = <cnp.uint32_t>n_edge 

    for i in range(n_edge):
        col  = <size_t>Aj[i]
        dest = <size_t>Bp[col]
        Bi[dest] = Ai[i]
        Bx[dest] = Ax[i]
        Bp[col] += 1

    last = 0
    for i in range(n_vert + 1):
        temp = Bp[i]
        Bp[i] = last
        last = temp


""" 
An implementation of Spiess and Florian's hyperpath generating algorithm.

reference: Spiess, H. and Florian, M. (1989). Optimal strategies: A new 
assignment model for transit networks. Transportation Research Part B 23(2), 
83-102.
    
author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""

import numpy as np
cimport numpy as cnp

from edsger.commons import DTYPE_PY, DTYPE_INF_PY
from edsger.commons cimport (
    DTYPE_INF, MIN_FREQ, UNLABELED, SCANNED, DTYPE_t, ElementState)
cimport edsger.pq_4ary_dec_0b as pq  # priority queue



cpdef void compute_SF_in(
    cnp.uint32_t[::1] csc_indptr,
    cnp.uint32_t[::1] csc_edge_idx,
    DTYPE_t[::1] c_a_vec,
    DTYPE_t[::1] f_a_vec,
    cnp.uint32_t[::1] tail_indices,
    cnp.uint32_t[::1] head_indices,
    cnp.uint32_t[::1] demand_indices,
    DTYPE_t[::1] demand_values,
    DTYPE_t[::1] v_a_vec,
    DTYPE_t[::1] u_i_vec,
    int vertex_count,
    int dest_vert_index,
):

    cdef:
        int edge_count = tail_indices.shape[0]

    # initialization
    u_i_vec[<size_t>dest_vert_index] = 0.0

    # vertex properties
    f_i_vec = np.zeros(vertex_count, dtype=DTYPE_PY)  # vertex frequency (inverse of the maximum delay)
    u_j_c_a_vec = DTYPE_INF_PY * np.ones(edge_count, dtype=DTYPE_PY)    
    v_i_vec = np.zeros(vertex_count, dtype=DTYPE_PY)  # vertex volume
    
    # edge properties
    h_a_vec = np.zeros(edge_count, dtype=bool)  # edge belonging to hyperpath

    # first pass #
    # ---------- #

    _SF_in_first_pass_full(
        csc_indptr,
        csc_edge_idx,
        c_a_vec,
        f_a_vec,
        tail_indices,
        u_i_vec,
        f_i_vec,
        u_j_c_a_vec,
        h_a_vec,
        dest_vert_index
    )

    # second pass #
    # ----------- #

    cdef:
        DTYPE_t u_r, v_a_new, v_i, u_i
        size_t i, h_a_count
        cnp.uint32_t vert_idx 

    v_i_vec = np.zeros(vertex_count, dtype=DTYPE_PY)  # vertex volume

    u_r = DTYPE_INF_PY
    for i, vert_idx in enumerate(demand_indices):

        v_i_vec[<size_t>vert_idx] = demand_values[i]
        u_i = u_i_vec[<size_t>vert_idx]

        if u_i < u_r:

            u_r = u_i

    # if the destination can be reached from any of the origins
    if u_r < DTYPE_INF_PY:

        # make sure f_i values are not zero
        f_i_vec = np.where(
            f_i_vec < MIN_FREQ_PY, MIN_FREQ_PY, f_i_vec
        )

        # sort the links with descreasing order of u_j + c_a
        h_a_count = h_a_vec.sum()
        masked_a = np.ma.array(-u_j_c_a_vec, mask=~h_a_vec)
        edge_indices = np.argsort(masked_a).astype(np.uint32)

        _SF_in_second_pass(
            edge_indices,
            tail_indices,
            head_indices,
            v_i_vec,
            v_a_vec,
            f_i_vec,
            f_a_vec,
            h_a_count
        )


cdef void _SF_in_first_pass_full(
    cnp.uint32_t[::1] csc_indptr, 
    cnp.uint32_t[::1] csc_edge_idx,
    DTYPE_t[::1] c_a_vec,
    DTYPE_t[::1] f_a_vec,
    cnp.uint32_t[::1] tail_indices,
    DTYPE_t[::1] u_i_vec,
    DTYPE_t[::1] f_i_vec,
    DTYPE_t[::1] u_j_c_a_vec,
    cnp.uint8_t[::1] h_a_vec,
    int dest_vert_index,
) nogil:
    """All vertices are visited."""

    cdef:
        int edge_count = tail_indices.shape[0]
        pq.PriorityQueue pqueue
        ElementState edge_state
        size_t i, edge_idx, tail_vert_idx
        DTYPE_t u_j_c_a, u_i, f_i, beta, u_i_new, f_a

    # initialization of the heap elements 
    # all nodes have INFINITY key and UNLABELED state
    pq.init_pqueue(&pqueue, <size_t>edge_count, <size_t>edge_count)

    # only the incoming edges of the target vertex are inserted into the 
    # priority queue
    for i in range(<size_t>csc_indptr[<size_t>dest_vert_index], 
        <size_t>csc_indptr[<size_t>(dest_vert_index + 1)]):
        edge_idx = csc_edge_idx[i]
        pq.insert(&pqueue, edge_idx, c_a_vec[edge_idx])
        u_j_c_a_vec[edge_idx] = c_a_vec[edge_idx]

    # first pass
    while pqueue.size > 0:

        edge_idx = pq.extract_min(&pqueue)
        u_j_c_a = pqueue.Elements[edge_idx].key
        tail_vert_idx = <size_t>tail_indices[edge_idx]
        u_i = u_i_vec[tail_vert_idx]

        if u_i >= u_j_c_a:

            f_i = f_i_vec[tail_vert_idx]

            # compute the beta coefficient
            if (u_i < DTYPE_INF) | (f_i > 0.0):

                beta = f_i * u_i

            else:

                beta = 1.0

            # update u_i
            f_a = f_a_vec[edge_idx]
            u_i_new = (beta + f_a * u_j_c_a) / (f_i + f_a)
            u_i_vec[tail_vert_idx] = u_i_new

            # update f_i
            f_i_vec[tail_vert_idx] = f_i + f_a

            # add the edge to hyperpath
            h_a_vec[edge_idx] = 1

        else:

            u_i_new = u_i

        # loop on incoming edges
        for i in range(<size_t>csc_indptr[tail_vert_idx], 
            <size_t>csc_indptr[tail_vert_idx + 1]):

            edge_idx = csc_edge_idx[i]
            edge_state = pqueue.Elements[edge_idx].state

            if edge_state != SCANNED:

                # u_j of current edge = u_i of outgoing edge
                u_j_c_a = u_i_new + c_a_vec[edge_idx]

                if edge_state == UNLABELED:

                    pq.insert(&pqueue, edge_idx, u_j_c_a)
                    u_j_c_a_vec[edge_idx] = u_j_c_a 

                elif (pqueue.Elements[edge_idx].key > u_j_c_a):

                    pq.decrease_key(&pqueue, edge_idx, u_j_c_a)
                    u_j_c_a_vec[edge_idx] = u_j_c_a

    pq.free_pqueue(&pqueue)


cdef void _SF_in_second_pass(
    cnp.uint32_t[::1] edge_indices,
    cnp.uint32_t[::1] tail_indices,
    cnp.uint32_t[::1] head_indices,
    DTYPE_t[::1] v_i_vec,
    DTYPE_t[::1] v_a_vec,
    DTYPE_t[::1] f_i_vec,
    DTYPE_t[::1] f_a_vec,
    size_t h_a_count
) nogil:

    cdef:
        size_t i, edge_idx, vert_idx
        DTYPE_t v_i, f_i, f_a, v_a_new

    for i in range(h_a_count):

        edge_idx = <size_t>edge_indices[i]
        vert_idx = <size_t>tail_indices[edge_idx]

        v_i = v_i_vec[vert_idx]
        f_i = f_i_vec[vert_idx]
        f_a = f_a_vec[edge_idx]

        # update v_a
        v_a_new = v_i * f_a / f_i
        v_a_vec[edge_idx] = v_a_new
        v_i_vec[<size_t>head_indices[edge_idx]] += v_a_new


# ============================================================================ #
# tests                                                                        #
# ============================================================================ #


from edsger.commons import MIN_FREQ_PY, INF_FREQ_PY

cpdef compute_SF_in_01():
    """  
    Single edge network.

    This network has 1 edge and 2 vertices.
    """

    volume = 1.0

    f_a = MIN_FREQ_PY

    csc_indptr = np.array([0, 0, 1], dtype=np.uint32)
    csc_edge_idx = np.array([0], dtype=np.uint32)
    c_a_vec = np.array([1.0], dtype=DTYPE_PY)
    f_a_vec = np.array([f_a], dtype=DTYPE_PY)
    v_a_vec = np.array([0.0], dtype=DTYPE_PY)
    tail_indices = np.array([0], dtype=np.uint32)
    head_indices = np.array([1], dtype=np.uint32)
    demand_indices = np.array([0], dtype=np.uint32)
    demand_values = np.array([volume], dtype=DTYPE_PY)
    vertex_count = 2
    u_i_vec = DTYPE_INF_PY * np.ones(vertex_count, dtype=DTYPE_PY)
    dest_vert_index = 1

    compute_SF_in(
        csc_indptr,  
        csc_edge_idx,
        c_a_vec,
        f_a_vec,
        tail_indices,
        head_indices,
        demand_indices,
        demand_values,
        v_a_vec,
        u_i_vec,
        vertex_count,
        dest_vert_index,
    )

    assert v_a_vec[0] == volume
    assert v_a_vec.shape[0] == 1

    f_a = INF_FREQ_PY
    f_a_vec = np.array([f_a], dtype=DTYPE_PY)
    u_i_vec = DTYPE_INF_PY * np.ones(vertex_count, dtype=DTYPE_PY)

    compute_SF_in(
        csc_indptr,  
        csc_edge_idx,
        c_a_vec,
        f_a_vec,
        tail_indices,
        head_indices,
        demand_indices,
        demand_values,
        v_a_vec,
        u_i_vec,
        vertex_count,
        dest_vert_index,
    )

    assert v_a_vec[0] == volume
    assert v_a_vec.shape[0] == 1


cpdef compute_SF_in_02():
    """  
    Two edges network.

    This network has 2 edges and 2 vertices.
    """

    volume = 1.0

    csc_indptr = np.array([0, 0, 2], dtype=np.uint32)
    csc_edge_idx = np.array([0, 1], dtype=np.uint32)
    c_a_vec = np.array([1.0, 1.0], dtype=DTYPE_PY)
    f_a_vec = np.array([2.0, 1.0], dtype=DTYPE_PY)
    v_a_vec = np.array([0.0, 0.0], dtype=DTYPE_PY)
    tail_indices = np.array([0, 0], dtype=np.uint32)
    head_indices = np.array([1, 1], dtype=np.uint32)
    demand_indices = np.array([0], dtype=np.uint32)
    demand_values = np.array([volume], dtype=DTYPE_PY)
    vertex_count = 2
    u_i_vec = DTYPE_INF_PY * np.ones(vertex_count, dtype=DTYPE_PY)
    dest_vert_index = 1

    compute_SF_in(
        csc_indptr,  
        csc_edge_idx,
        c_a_vec,
        f_a_vec,
        tail_indices,
        head_indices,
        demand_indices,
        demand_values,
        v_a_vec,
        u_i_vec,
        vertex_count,
        dest_vert_index,
    )

    assert v_a_vec[0] == 2. / 3.
    assert v_a_vec[1] == 1. / 3.
    assert v_a_vec.shape[0] == 2


"""
Run Dijkstra's algorithm (SSSP) on DIMACS networks.

Example :
> python dijkstra_dimacs.py -n USA -r 4 -c True
"""

import os
import sys
from argparse import ArgumentParser
from time import perf_counter

import numpy as np
import pandas as pd
from loguru import logger
from scipy.sparse import coo_array
from scipy.sparse.csgraph import dijkstra

from edsger.path import Dijkstra


logger.remove()
fmt = (
    "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> |"
    + " <level>{message}</level>"
)
logger.add(sys.stderr, format=fmt)

parser = ArgumentParser(description="Command line interface to perf_01.py")
parser.add_argument(
    "-d",
    "--dir",
    dest="data_dir",
    help="Data folder with network sub-folders",
    metavar="TXT",
    type=str,
    required=False,
    default="/home/francois/Data/DIMACS_road_networks/",
)
parser.add_argument(
    "-n",
    "--network",
    dest="network_name",
    help="network name, must be 'NY', 'BAY', 'COL', 'FLA', 'NW', "
    + "'NE', 'CAL', 'LKS', 'E', 'W', 'CTR', 'USA'",
    metavar="TXT",
    type=str,
    required=True,
)
parser.add_argument(
    "-l",
    "--library",
    dest="library_name",
    help='library name, must be "E" (Edsger), "GT" (graph-tool), "NK" (NetworKit)',
    metavar="TXT",
    type=str,
    required=False,
    default="E",
)
parser.add_argument(
    "-f",
    "--from",
    dest="idx_from",
    help="source vertex index",
    metavar="INT",
    type=int,
    required=False,
    default=1000,
)
parser.add_argument(
    "-r",
    "--repeat",
    dest="repeat",
    help="repeat",
    metavar="INT",
    type=int,
    required=False,
    default=1,
)
parser.add_argument(
    "-c",
    "--check_result",
    dest="check_result",
    help="check the resulting path lengths aginst SciPy",
    action="store_true",
)
parser.add_argument(
    "-q",
    "--heap_length_ratio",
    dest="heap_length_ratio",
    help="heap length ratio (0, 1]",
    metavar="FLOAT",
    type=float,
    required=False,
    default=1.0,
)
args = parser.parse_args()
data_dir = args.data_dir
data_dir = os.path.abspath(data_dir)
reg = args.network_name
reg = reg.upper()
idx_from = args.idx_from
repeat = args.repeat
check_result = args.check_result
lib = args.library_name.upper()
heap_length_ratio = args.heap_length_ratio

# lib name check
assert lib in ["E", "GT", "NK"]

data_dir_found = os.path.exists(data_dir)
if data_dir_found:
    logger.info(f"data dir : {data_dir}")
else:
    logger.critical(f"data dir '{data_dir}' not found")
    sys.exit()

# network name check
regions_usa = [
    "NY",
    "BAY",
    "COL",
    "FLA",
    "NW",
    "NE",
    "CAL",
    "LKS",
    "E",
    "W",
    "CTR",
    "USA",
]
if reg in regions_usa:
    logger.info(f"region : {reg}")
else:
    logger.critical(f"region '{reg}' invalid")
    sys.exit()

if isinstance(idx_from, int) and (idx_from >= 0):
    logger.info(f"idx_from : {idx_from}")
else:
    logger.critical(f"invalid value '{idx_from}' for idx_from")
    sys.exit()

if isinstance(repeat, int) and (repeat > 0):
    logger.info(f"repeat : {repeat}")
else:
    logger.critical(f"invalid value'{repeat}' for repeat")
    sys.exit()

if isinstance(check_result, bool):
    logger.info(f"check result : {check_result}")
else:
    logger.critical(f"invalid value '{check_result}' for check_result")
    sys.exit()

# locate the parquet file
network_file_path = os.path.join(data_dir, f"{reg}/USA-road-t.{reg}.gr.parquet")
network_file_found = os.path.exists(network_file_path)
if network_file_found:
    logger.info(f"network file path : {network_file_path}")
else:
    logger.critical(f"network file path '{network_file_path}' not found")
    sys.exit()

# load the network into a Pandas dataframe
edges = pd.read_parquet(network_file_path)
edges.rename(columns={"source": "tail", "target": "head", "tt": "weight"}, inplace=True)
edge_count = len(edges)
vertex_count = edges[["tail", "head"]].max().max() + 1
logger.info(f"{edge_count} edges and {vertex_count} vertices")

if lib == "E":
    # Edsger
    # ------

    logger.info("Edsger init")
    edges[["tail", "head"]] = edges[["tail", "head"]].astype(np.uint32)
    sp = Dijkstra(edges, orientation="out", check_edges=False)

    # SSSP

    results = []
    logger.info("Edsger Run")
    for i in range(repeat):
        d = {}

        start = perf_counter()

        dist_matrix = sp.run(
            vertex_idx=idx_from, return_inf=True, heap_length_ratio=heap_length_ratio
        )

        end = perf_counter()
        elapsed_time = end - start
        logger.info(
            f"Edsger Dijkstra {i+1}/{repeat} - Elapsed time: {elapsed_time:8.4f} s"
        )

        d = {
            "library": "edsger",
            "network": reg,
            "trial": i,
            "elapsed_time": elapsed_time,
        }
        results.append(d)

    df = pd.DataFrame.from_records(results)
    logger.info(f"Edsger min elapsed time : {df.elapsed_time.min():8.4f} s")

elif lib == "GT":
    # graph-tool
    # ----------

    logger.info("graph-tool init")

    import graph_tool as gt
    from graph_tool import topology

    # create the graph
    g = gt.Graph(directed=True)

    # create the vertices
    g.add_vertex(vertex_count)

    # create the edges
    g.add_edge_list(edges[["tail", "head"]].values)

    # edge property for the travel time
    eprop_t = g.new_edge_property("float")
    g.edge_properties["t"] = eprop_t  # internal property
    g.edge_properties["t"].a = edges["weight"].values

    # SSSP

    results = []
    logger.info("graph-tool Run")
    for i in range(repeat):
        d = {}

        start = perf_counter()

        dist = gt.topology.shortest_distance(
            g,
            source=g.vertex(idx_from),
            weights=g.ep.t,
            negative_weights=False,
            directed=True,
        )
        dist_matrix = dist.a

        end = perf_counter()
        elapsed_time = end - start
        logger.info(
            f"graph-tool Dijkstra {i+1}/{repeat} - Elapsed time: {elapsed_time:8.4f} s"
        )

        d = {
            "library": "graph-tool",
            "network": reg,
            "trial": i,
            "elapsed_time": elapsed_time,
        }
        results.append(d)

    df = pd.DataFrame.from_records(results)
    logger.info(f"graph-tool min elapsed time : {df.elapsed_time.min():8.4f} s")

elif lib == "NK":
    # NetworKit
    # ---------

    import networkit as nk

    nk_file_format = nk.graphio.Format.NetworkitBinary
    networkit_file_path = os.path.join(
        data_dir, f"{reg}/USA-road-t.{reg}.gr.NetworkitBinary"
    )

    if os.path.exists(networkit_file_path):
        g = nk.graphio.readGraph(networkit_file_path, nk_file_format)

    else:
        g = nk.Graph(n=vertex_count, weighted=True, directed=True, edgesIndexed=False)

        for row in edges.itertuples():
            g.addEdge(row.tail, row.head, w=row.weight)

        nk.graphio.writeGraph(g, networkit_file_path, nk_file_format)

    nk_dijkstra = nk.distance.Dijkstra(
        g, idx_from, storePaths=False, storeNodesSortedByDistance=False
    )

    # SSSP

    results = []
    logger.info("NetworKit Run")
    for i in range(repeat):
        d = {}

        start = perf_counter()

        nk_dijkstra.run()
        dist_matrix = np.asarray(nk_dijkstra.getDistances(asarray=True))
        dist_matrix = np.where(dist_matrix >= 1.79769313e308, np.inf, dist_matrix)

        end = perf_counter()
        elapsed_time = end - start
        logger.info(
            f"NetworKit Dijkstra {i+1}/{repeat} - Elapsed time: {elapsed_time:8.4f} s"
        )

        d = {
            "library": "NetworKit",
            "network": reg,
            "trial": i,
            "elapsed_time": elapsed_time,
        }
        results.append(d)

    df = pd.DataFrame.from_records(results)
    logger.info(f"NetworKit min elapsed time : {df.elapsed_time.min():8.4f} s")


if check_result:
    logger.info(f"result check")

    # SciPy
    logger.info("SciPy init")
    data = edges["weight"].values
    row = edges["tail"].values
    col = edges["head"].values
    graph_coo = coo_array((data, (row, col)), shape=(vertex_count, vertex_count))
    graph_csr = graph_coo.tocsr()

    logger.info("SciPy run")
    start = perf_counter()
    dist_matrix_ref = dijkstra(
        csgraph=graph_csr,
        directed=True,
        indices=idx_from,
        return_predecessors=False,
    )
    end = perf_counter()
    elapsed_time = end - start

    logger.info(f"SciPy Dijkstra - Elapsed time: {elapsed_time:8.4f} s")
    logger.info(f"isinf : {np.isinf(dist_matrix_ref).any()}")
    logger.info(
        f"allclose : {np.allclose(dist_matrix_ref, dist_matrix, equal_nan=True)}"
    )


logger.info("exit")



"""setup python file for edsger.
"""

import numpy as np
from Cython.Build import cythonize
from setuptools import Extension, setup

extra_compile_args = ["-Ofast"]

extensions = [
    Extension(
        "edsger.commons",
        ["src/edsger/commons.pyx"],
        extra_compile_args=extra_compile_args,
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    ),
    Extension(
        "edsger.pq_4ary_dec_0b",
        ["src/edsger/pq_4ary_dec_0b.pyx"],
        extra_compile_args=extra_compile_args,
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    ),
    Extension(
        "edsger.dijkstra",
        ["src/edsger/dijkstra.pyx"],
        extra_compile_args=extra_compile_args,
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    ),
    Extension(
        "edsger.spiess_florian",
        ["src/edsger/spiess_florian.pyx"],
        extra_compile_args=extra_compile_args,
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    ),
    Extension(
        "edsger.star",
        ["src/edsger/star.pyx"],
        extra_compile_args=extra_compile_args,
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    ),
    Extension(
        "edsger.path_tracking",
        ["src/edsger/path_tracking.pyx"],
        extra_compile_args=extra_compile_args,
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    ),
]

with open("requirements.txt") as fp:
    install_requires = fp.read().strip().split("\n")


def setup_package():
    setup(
        ext_modules=cythonize(
            extensions,
            compiler_directives={
                "language_level": "3",
                "boundscheck": False,
                "wraparound": False,
                "embedsignature": False,
                "cdivision": True,
                "initializedcheck": False,
            },
            include_path=["src/edsger/"],
        ),
        install_requires=install_requires,
        include_dirs=[np.get_include()],
    )


if __name__ == "__main__":
    setup_package()



""" Tests for spiess_florian.pyx.

py.test tests/test_spiess_florian.py

author : Francois Pacull
copyright : Architecture & Performance
email: francois.pacull@architecture-performance.fr
license : MIT
"""


from edsger.spiess_florian import *


def test_compute_SF_in_01():
    compute_SF_in_01()


def test_compute_SF_in_02():
    compute_SF_in_02()
